from . import core, mixin
from .decorators import _inplace_enabled, _inplace_enabled_define_and_cleanup


class DomainTopology(
    mixin.NetCDFVariable,
    mixin.Topology,
    mixin.PropertiesData,
    mixin.Files,
    core.DomainTopology,
):
    """A domain topology construct of the CF data model.

    A domain topology construct defines the geospatial topology of
    cells arranged in two or three dimensions in real space but
    indexed by a single (discrete) domain axis construct, and at most
    one domain topology construct may be associated with any such
    domain axis. The topology describes topological relationships
    between the cells - spatial relationships which do not depend on
    the cell locations - and is represented by an undirected graph,
    i.e. a mesh in which pairs of nodes are connected by links. Each
    node has a unique arbitrary identity that is independent of its
    spatial location, and different nodes may be spatially co-located.

    The topology may only describe cells that have a common spatial
    dimensionality, one of:

    * Point: A point is zero-dimensional and has no boundary vertices.
    * Edge: An edge is one-dimensional and corresponds to a line
            connecting two boundary vertices.
    * Face: A face is two-dimensional and corresponds to a surface
            enclosed by a set of edges.

    Each type of cell implies a restricted topology for which only
    some kinds of mesh are allowed. For point cells, every node
    corresponds to exactly one cell; and two cells have a topological
    relationship if and only if their nodes are connected by a mesh
    link. For edge and face cells, every node corresponds to a
    boundary vertex of a cell; the same node can represent vertices in
    multiple cells; every link in the mesh connects two cell boundary
    vertices; and two cells have a topological relationship if and
    only if they share at least one node.

    A domain topology construct contains an array defining the mesh,
    and properties to describe it. There must be a property indicating
    the spatial dimensionality of the cells. The array values comprise
    the node identities, and all array elements that refer to the same
    node must contain the same value, which must differ from any other
    value in the array. The array spans the domain axis construct and
    also has a ragged dimension, whose function depends on the spatial
    dimensionality of the cells.

    For each point cell, the first element along the ragged dimension
    contains the node identity of the cell, and the following elements
    contain in arbitrary order the identities of all the cells to
    which it is connected by a mesh link.

    For each edge or face cell, the elements along the ragged
    dimension contain the node identities of the boundary vertices of
    the cell, in the same order that the boundary vertices are stored
    by the auxiliary coordinate constructs. Each boundary vertex
    except the last is connected by a mesh link to the next vertex
    along the ragged dimension, and the last vertex is connected to
    the first.

    When a domain topology construct is present it is considered to be
    definitive and must be used in preference to the topology implied
    by inspection of any other constructs, which is not guaranteed to
    be the same.

    In CF-netCDF a domain topology construct can only be provided for
    a UGRID mesh topology variable. The information in the construct
    array is supplied by the UGRID "edge_nodes_connectivity" variable
    (for edge cells) or "face_nodes_connectivity" variable (for face
    cells). The topology for node cells may be provided by any of
    these three UGRID variables. The integer indices contained in the
    UGRID variable may be used as the mesh node identities, although
    the CF data model attaches no significance to the values other
    than the fact that some values are the same as others. The spatial
    dimensionality property is provided by the "location" attribute of
    a variable that references the UGRID mesh topology variable,
    i.e. a data variable or a UGRID location index set variable.

    A single UGRID mesh topology defines multiple domain constructs
    and defines how they relate to each other. For instance, when
    "face_node_connectivity" and "edge_node_connectivity" variables
    are both present there are three implied domain constructs - one
    each for face, edge and point cells - all of which have the same
    mesh and so are explicitly linked (e.g. it is known which edge
    cells define each face cell). The CF data model has no mechanism
    for explicitly recording such relationships between multiple
    domain constructs, however whether or not two domains have the
    same mesh may be reliably determined by inspection, thereby
    allowing the creation of netCDF datasets containing UGRID mesh
    topology variables.

    The restrictions on the type of mesh that may be used with a given
    cell spatial dimensionality excludes some meshes which can be
    described by an undirected graph, but is consistent with UGRID
    encoding within CF-netCDF. UGRID also describes meshes for
    three-dimensional volume cells that correspond to a volume
    enclosed by a set of faces, but how the nodes relate to volume
    boundary vertices is undefined and so volume cells are currently
    omitted from the CF data model.

    See CF Appendix I "The CF Data Model".

    **NetCDF interface**

    The netCDF variable name of the UGRID mesh topology variable may
    be accessed with the `nc_set_variable`, `nc_get_variable`,
    `nc_del_variable`, and `nc_has_variable` methods.

    {{netCDF HDF5 chunks}}

    .. versionadded:: (cfdm) 1.11.0.0

    """

    def __init__(
        self,
        cell=None,
        properties=None,
        data=None,
        source=None,
        copy=True,
        _use_data=True,
    ):
        """**Initialisation**

        :Parameters:

            {{init cell: `str`, optional}}

            {{init properties: `dict`, optional}}

                *Parameter example:*
                  ``properties={'long_name': 'face to node mapping'}``

            {{init data: data_like, optional}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            cell=cell,
            properties=properties,
            data=data,
            source=source,
            copy=copy,
            _use_data=_use_data,
        )

        self._initialise_netcdf(source)
        self._initialise_original_filenames(source)

    def creation_commands(
        self,
        representative_data=False,
        namespace=None,
        indent=0,
        string=True,
        name="c",
        data_name="data",
        header=True,
    ):
        """Returns the commands to create the domain topology construct.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `{{package}}.Data.creation_commands`,
                     `{{package}}.Field.creation_commands`

        :Parameters:

            {{representative_data: `bool`, optional}}

            {{namespace: `str`, optional}}

            {{indent: `int`, optional}}

            {{string: `bool`, optional}}

            {{name: `str`, optional}}

            {{data_name: `str`, optional}}

            {{header: `bool`, optional}}

        :Returns:

            {{returns creation_commands}}

        """
        out = super().creation_commands(
            representative_data=representative_data,
            indent=0,
            namespace=namespace,
            string=False,
            name=name,
            data_name=data_name,
            header=header,
        )

        cell = self.get_cell(None)
        if cell is not None:
            out.append(f"{name}.set_cell({cell!r})")

        if string:
            indent = " " * indent
            out[0] = indent + out[0]
            out = ("\n" + indent).join(out)

        return out

    def dump(
        self,
        display=True,
        _omit_properties=None,
        _key=None,
        _level=0,
        _title=None,
        _axes=None,
        _axis_names=None,
    ):
        """A full description of the domain topology construct.

        Returns a description of all properties, including those of
        components, and provides selected values of all data arrays.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default the description is printed.

        :Returns:

            {{returns dump}}

        """
        if _title is None:
            _title = "Domain Topology: " + self.identity(default="")

        return super().dump(
            display=display,
            _key=_key,
            _omit_properties=_omit_properties,
            _level=_level,
            _title=_title,
            _axes=_axes,
            _axis_names=_axis_names,
        )

    def identity(self, default=""):
        """Return the canonical identity.

        By default the identity is the first found of the following:

        * The cell type, preceded by ``'cell:'``.
        * The ``standard_name`` property.
        * The ``cf_role`` property, preceded by 'cf_role='.
        * The ``long_name`` property, preceded by 'long_name='.
        * The netCDF variable name, preceded by 'ncvar%'.
        * The value of the default parameter.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `identities`

        :Parameters:

            default: optional
                If no identity can be found then return the value of the
                default parameter.

        :Returns:

                The identity.

        """
        n = self.get_cell(None)
        if n is not None:
            return f"cell:{n}"

        n = self.get_property("standard_name", None)
        if n is not None:
            return n

        for prop in ("cf_role", "long_name"):
            n = self.get_property(prop, None)
            if n is not None:
                return f"{prop}={n}"

        n = self.nc_get_variable(None)
        if n is not None:
            return f"ncvar%{n}"

        return default

    def identities(self, generator=False, **kwargs):
        """Return all possible identities.

        The identities comprise:

        * The cell type type, preceded by ``'cell:'``.
        * The ``standard_name`` property.
        * All properties, preceded by the property name and a colon,
          e.g. ``'long_name:Air temperature'``.
        * The netCDF variable name, preceded by ``'ncvar%'``.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `identity`

        :Parameters:

            generator: `bool`, optional
                If True then return a generator for the identities,
                rather than a list.

            kwargs: optional
                Additional configuration parameters that may be used
                by subclasses.

        :Returns:

            `list` or generator
                The identities.

        """
        n = self.get_cell(None)
        if n is not None:
            pre = ((f"cell:{n}",),)
            pre0 = kwargs.pop("pre", None)
            if pre0:
                pre = tuple(pre0) + pre

            kwargs["pre"] = pre

        return super().identities(generator=generator, **kwargs)

    @_inplace_enabled(default=False)
    def normalise(
        self, start_index=0, remove_empty_columns=False, inplace=False
    ):
        """Normalise the data values.

        Normalised data is in a form that is suitable for creating a
        CF-netCDF UGRID "edge_node_connectivity" or
        "face_node_connectivity" variable.

        Normalisation does not change the logical content of the
        data. It converts the data so that the set of values comprises
        all of the integers in the range ``[0, N-1]`` (if the
        *start_index* parameter is ``0``), or ``[1, N]`` (if
        *start_index* is ``1``), where ``N`` is the number of mesh
        nodes.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            start_index: `int`, optional
                The start index for the data values in the normalised
                data. Must be ``0`` (the default) or ``1`` for zero-
                or one-based indices respectively.

            remove_empty_columns: `bool`, optional
                If True then remove any array columns that are
                entirely missing data. By default these are retained.

            {{inplace: `bool`, optional}}

        :Returns:

            `{{class}}` or `None`
                The normalised domain topology construct, or `None` if
                the operation was in-place.

        **Examples*

        Face cells (similarly for edge cells):

        >>> data = {{package}}.Data(
        ...   [[1, 4, 5, 2], [4, 10, 1, -99], [122, 123, 106, 105]],
        ...   mask_value=-99
        ... )
        >>> d = {{package}}.{{class}}(cell='face', data=data)
        >>> print(d.array)
        [[1 4 5 2]
         [4 10 1 --]
         [122 123 106 105]]
        >>> d0 = d.normalise()
        >>> print(c.array)
        [[0 2 3 1]
         [2 4 0 --]
         [7 8 6 5]]
        >>> (d0.array == d.normalise().array).all()
        True
        >>> d1 = d.normalise(start_index=1)
        >>> print(d1.array)
        [[1 3 4 2]
         [3 5 1 --]
         [8 9 7 6]]
        >>> (d1.array == d0.array + 1).all()
        True

        Point cells:

        >>> data = {{package}}.Data(
        ...   [[4, 1, 10, 125], [1, 4, -99, -99], [125, 4, -99, -99]],
        ...   mask_value=-99
        ... )
        >>> d = {{package}}.{{class}}(cell='point', data=data)
        >>> print(d.array)
        [[4 1 10 125]
         [1 4 -- --]
         [125 4 -- --]]
        >>> print(d.normalise().array)
        [[0 1 2]
         [1 0 --]
         [2 0 --]]
        >>> print(d.normalise(start_index=1).array)
        [[1 2 3]
         [2 1 --]
         [3 1 --]]

        """
        import numpy as np

        if start_index not in (0, 1):
            raise ValueError("The 'start_index' parameter must be 0 or 1")

        cell = self.get_cell(None)
        if cell is None:
            raise ValueError(
                f"Can't normalise {self.__class__.__name__} with unknown "
                "cell type"
            )

        d = _inplace_enabled_define_and_cleanup(self)
        data = d.array

        if cell in ("edge", "face"):
            # Normalise node ids for edge or face cells.
            mask = np.ma.getmaskarray(data)
            n, b = np.where(~mask)
            data[n, b] = np.unique(data[n, b], return_inverse=True)[1]

            if remove_empty_columns:
                # Discard columns that are all missing data
                data = self._remove_empty_columns(data)

            if start_index:
                data += 1
        elif cell == "point":
            # Normalise cell ids for point cells
            data = self._normalise_cell_ids(
                data, start_index, remove_empty_columns
            )
        else:
            raise ValueError(f"Can't normalise: Unknown cell type {cell!r}")

        d.set_data(data, copy=False)
        return d



from .parameters import Parameters


class ParametersDomainAncillaries(Parameters):
    """Mixin to collect named parameters and domain ancillaries.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(
        self, parameters=None, domain_ancillaries=None, source=None, copy=True
    ):
        """**Initialisation**

        :Parameters:

            parameters: `dict`, optional
               Set parameters. The dictionary keys are term names,
               with corresponding parameter values.

               Parameters may also be set after initialisation with
               the `set_parameters` and `set_parameter` methods.

               *Parameter example:*
                 ``parameters={'earth_radius': 6371007.}``

            domain_ancillaries: `dict`, optional
               Set references to domain ancillary constructs. The
               dictionary keys are term names, with corresponding
               domain ancillary construct keys.

               Domain ancillaries may also be set after initialisation
               with the `set_domain_ancillaries` and
               `set_domain_ancillary` methods.

               *Parameter example:*
                 ``domain_ancillaries={'orog': 'domainancillary2'}``

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(parameters=parameters, source=source, copy=copy)

        self._set_component("domain_ancillaries", {}, copy=False)

        if source:
            try:
                domain_ancillaries = source.domain_ancillaries()
            except AttributeError:
                domain_ancillaries = None

        if domain_ancillaries is None:
            domain_ancillaries = {}

        self.set_domain_ancillaries(domain_ancillaries, copy=False)

    def clear_domain_ancillaries(self):
        """Remove all domain_ancillaries.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_domain_ancillary`, `domain_ancillaries`,
                     `set_domain_ancillaries`

        :Returns:

            `dict`
                The domain ancillaries that have been removed.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.domain_ancillaries()
        {}
        >>> d = {'a': 'domainancillary0',
        ...      'b': 'domainancillary1',
        ...      'orog': 'domainancillary2'}
        >>> f.set_domain_ancillaries(d)
        >>> f.domain_ancillaries()
        {'a': 'domainancillary0',
         'b': 'domainancillary1',
         'orog': 'domainancillary2'}

        >>> old = f.clear_domain_ancillaries()
        >>> f.domain_ancillaries()
        {}
        >>> old
        {'a': 'domainancillary0',
         'b': 'domainancillary1',
         'orog': 'domainancillary2'}

        """
        out = self._get_component("domain_ancillaries")
        self._set_component("domain_ancillaries", {})
        return out.copy()

    def del_domain_ancillary(self, domain_ancillary, default=ValueError()):
        """Delete a domain ancillary.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `domain_ancillaries`, `get_domain_ancillary`,
                     `set_domain_ancillary`

        :Parameters:

            domain_ancillary: `str`
                The name of the domain ancillary to be deleted.

                *Parameter example:*
                   ``domain_ancillary='orog'``

            default: optional
                Return the value of the *default* parameter if the domain
                ancillary term has not been set.

                {{default Exception}}

        :Returns:

            `str`
                The removed domain ancillary key.

        **Examples**

        >>> c = {{package}}.{{class}}()
        >>> c.set_domain_ancillary('orog', 'domainancillary2')
        >>> c.has_domain_ancillary('orog')
        True
        >>> c.get_domain_ancillary('orog')
        'domainancillary2'
        >>> c.del_domain_ancillary('orog')
        'domainancillary2'
        >>> c.has_domain_ancillary('orog')
        False
        >>> print(c.del_domain_ancillary('orog', None))
        None
        >>> print(c.get_domain_ancillary('orog', None))
        None

        """
        try:
            return self._get_component("domain_ancillaries").pop(
                domain_ancillary
            )
        except KeyError:
            if default is None:
                return

            return self._default(
                default,
                f"{self.__class__.__name__!r} has no {domain_ancillary!r} domain ancillary",
            )

    def domain_ancillaries(self):
        """Return all domain_ancillaries.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `clear_domain_ancillaries`, `get_domain_ancillary`,
                     `has_domain_ancillaryr` `set_domain_ancillaries`

        :Returns:

            `dict`
                The domain ancillaries.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.domain_ancillaries()
        {}
        >>> d = {'a': 'domainancillary0',
        ...      'b': 'domainancillary1',
        ...      'orog': 'domainancillary2'}
        >>> f.set_domain_ancillaries(d)
        >>> f.domain_ancillaries()
        {'a': 'domainancillary0',
         'b': 'domainancillary1',
         'orog': 'domainancillary2'}

        >>> old = f.clear_domain_ancillaries()
        >>> f.domain_ancillaries()
        {}
        >>> old
        {'a': 'domainancillary0',
         'b': 'domainancillary1',
         'orog': 'domainancillary2'}

        """
        return self._get_component("domain_ancillaries").copy()

    def get_domain_ancillary(self, domain_ancillary, default=ValueError()):
        """Return a domain ancillary term.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_domain_ancillary`, `domain_ancillaries`,
                     `set_domain_ancillary`

        :Parameters:

            domain_ancillary: `str`
                The name of the term.

            default: optional
                Return the value of the *default* parameter if the domain
                ancillary term has not been set.

                {{default Exception}}

        :Returns:

                The domain ancillary construct key.

        **Examples**

        >>> c = {{package}}.{{class}}()
        >>> c.set_domain_ancillary('orog', 'domainancillary2')
        >>> c.has_domain_ancillary('orog')
        True
        >>> c.get_domain_ancillary('orog')
        'domainancillary2'
        >>> c.del_domain_ancillary('orog')
        'domainancillary2'
        >>> c.has_domain_ancillary('orog')
        False
        >>> print(c.del_domain_ancillary('orog', None))
        None
        >>> print(c.get_domain_ancillary('orog', None))
        None

        """
        try:
            return self._get_component("domain_ancillaries")[domain_ancillary]
        except KeyError:
            if default is None:
                return

            return self._default(
                default,
                f"{self.__class__.__name__!r} has no {domain_ancillary!r} domain ancillary",
            )

    def has_domain_ancillary(self, domain_ancillary):
        """Whether a domain ancillary has been set.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_domain_ancillary`, `domain_ancillaries`,
                     `has_domain_ancillary`, `set_domain_ancillary`

        :Parameters:

            domain_ancillary: `str`
                The name of the term.

        :Returns:

                The domain ancillary construct key.

        **Examples**

        >>> c = {{package}}.{{class}}()
        >>> c.set_domain_ancillary('orog', 'domainancillary2')
        >>> c.has_domain_ancillary('orog')
        True
        >>> c.get_domain_ancillary('orog')
        'domainancillary2'
        >>> c.del_domain_ancillary('orog')
        'domainancillary2'
        >>> c.has_domain_ancillary('orog')
        False
        >>> print(c.del_domain_ancillary('orog', None))
        None
        >>> print(c.get_domain_ancillary('orog', None))
        None

        """
        return domain_ancillary in self._get_component("domain_ancillaries")

    def set_domain_ancillaries(self, domain_ancillaries, copy=True):
        """Set domain_ancillaries.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `clear_domain_ancillaries`, `domain_ancillaries`,
                     `set_domain_ancillary`

        :Parameters:

            domain_ancillaries: `dict`
                Store the domain ancillaries from the dictionary supplied.

                *Parameter example:*
                  ``domain_ancillaries={'earth_radius': 6371007}``

            copy: `bool`, optional
                If False then any parameter values provided by the
                *domain_ancillaries* parameter are not copied before
                insertion. By default they are deep copied.

        :Returns:

            `None`

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.domain_ancillaries()
        {}
        >>> d = {'a': 'domainancillary0',
        ...      'b': 'domainancillary1',
        ...      'orog': 'domainancillary2'}
        >>> f.set_domain_ancillaries(d)
        >>> f.domain_ancillaries()
        {'a': 'domainancillary0',
         'b': 'domainancillary1',
         'orog': 'domainancillary2'}

        >>> old = f.clear_domain_ancillaries()
        >>> f.domain_ancillaries()
        {}
        >>> old
        {'a': 'domainancillary0',
         'b': 'domainancillary1',
         'orog': 'domainancillary2'}

        """
        self._get_component("domain_ancillaries").update(domain_ancillaries)

    def set_domain_ancillary(self, term, value, copy=True):
        """Set an domain ancillary-valued term.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_domain_ancillary`, `domain_ancillaries`,
                     `get_domain_ancillary`

        :Parameters:

            term: `str`
                The name of the term to be set.

            value:
                The value for the term.

            copy: `bool`, optional
                If True then set a deep copy of *value*.

        :Returns:

            `None`

        **Examples**

        >>> c = {{package}}.{{class}}()
        >>> c.set_domain_ancillary('orog', 'domainancillary2')
        >>> c.has_domain_ancillary('orog')
        True
        >>> c.get_domain_ancillary('orog')
        'domainancillary2'
        >>> c.del_domain_ancillary('orog')
        'domainancillary2'
        >>> c.has_domain_ancillary('orog')
        False
        >>> print(c.del_domain_ancillary('orog', None))
        None
        >>> print(c.get_domain_ancillary('orog', None))
        None

        """
        self._get_component("domain_ancillaries")[term] = value



from . import core, mixin


class Index(
    mixin.NetCDFVariable,
    mixin.NetCDFDimension,
    mixin.NetCDFSampleDimension,
    mixin.PropertiesData,
    mixin.Files,
    core.abstract.PropertiesData,
):
    """An index variable required to uncompress a ragged array.

    A collection of features stored using an indexed ragged array
    combines all features along a single dimension (the sample
    dimension) such that the values of each feature in the collection
    are interleaved.

    The information needed to uncompress the data is stored in an
    index variable that specifies the feature that each element of the
    sample dimension belongs to.

    **NetCDF interface**

    {{netCDF variable}}

    {{netCDF HDF5 chunks}}

    The name of the netCDF dimension spanned by the index variable's
    data (which does not correspond to a domain axis construct) may be
    accessed with the `nc_set_dimension`, `nc_get_dimension`,
    `nc_del_dimension`, and `nc_has_dimension` methods.

    The name of the netCDF sample dimension spanned by the compressed
    data (which does not correspond to a domain axis contract) may be
    accessed with the `nc_set_sample_dimension`,
    `nc_get_sample_dimension`, `nc_del_sample_dimension`, and
    `nc_has_sample_dimension` methods.

       .. note:: The netCDF sample dimension and the netCDF dimension
                 spanned by the index variable's data are should be
                 the same, unless the compressed data is an indexed
                 contiguous ragged array, in which case they must be
                 different.

    The name of the netCDF instance dimension (that is stored in the
    "instance_dimension" netCDF attribute) is accessed via the
    corresponding domain axis construct.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(
        self,
        properties=None,
        data=None,
        source=None,
        copy=True,
        _use_data=True,
    ):
        """**Initialisation**

        :Parameters:

            {{init properties: `dict`, optional}}

                *Parameter example:*
                  ``properties={'long_name': 'which station this obs is for'}``

            {{init data: data_like, optional}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            properties=properties,
            data=data,
            source=source,
            copy=copy,
            _use_data=_use_data,
        )

        self._initialise_netcdf(source)
        self._initialise_original_filenames(source)

    def dump(
        self,
        display=True,
        _key=None,
        _title=None,
        _create_title=True,
        _prefix="",
        _level=0,
        _omit_properties=None,
        _axes=None,
        _axis_names=None,
    ):
        """A full description of the index variable.

        Returns a description of all properties, including those of
        components, and provides selected values of all data arrays.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default the description is printed.

        :Returns:

            {{returns dump}}

        """
        if _create_title and _title is None:
            _title = "Index: " + self.identity(default="")

        return super().dump(
            display=display,
            _key=_key,
            _omit_properties=_omit_properties,
            _prefix=_prefix,
            _level=_level,
            _title=_title,
            _create_title=_create_title,
            _axes=_axes,
            _axis_names=_axis_names,
        )



import atexit
import datetime
import faulthandler
import os
import tempfile
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

n_tmpfiles = 1
tmpfiles = [
    tempfile.mkstemp("_test_geometry.nc", dir=os.getcwd())[1]
    for i in range(n_tmpfiles)
]
(tempfile,) = tmpfiles


def _remove_tmpfiles():
    """Remove temporary files created during tests."""
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass


atexit.register(_remove_tmpfiles)


VN = cfdm.CF()


class GeometryTest(unittest.TestCase):
    """Test the management of geometry cells."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.log_level('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

        self.geometry_1_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "geometry_1.nc"
        )
        self.geometry_2_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "geometry_2.nc"
        )
        self.geometry_3_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "geometry_3.nc"
        )
        self.geometry_4_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "geometry_4.nc"
        )
        self.geometry_interior_ring_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "geometry_interior_ring.nc",
        )
        self.geometry_interior_ring_file_2 = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "geometry_interior_ring_2.nc",
        )

    def test_node_count(self):
        """Test geometry coordinate node count variables."""
        f = cfdm.read(self.geometry_1_file, verbose=False)

        self.assertEqual(len(f), 2, "f = " + repr(f))
        for g in f:
            self.assertTrue(g.equals(g.copy(), verbose=1))
            self.assertEqual(len(g.auxiliary_coordinates()), 2)

        g = f[0]
        for axis in ("X", "Y"):
            coord = g.construct("axis=" + axis)
            self.assertTrue(coord.has_node_count(), "axis=" + axis)
            self.assertFalse(coord.has_part_node_count(), "axis=" + axis)
            self.assertFalse(coord.has_interior_ring(), "axis=" + axis)

        cfdm.write(f, tempfile, Conventions="CF-" + VN, verbose=False)

        f2 = cfdm.read(tempfile, verbose=False)
        self.assertEqual(len(f2), 2, "f2 = " + repr(f2))
        for a, b in zip(f, f2):
            self.assertTrue(a.equals(b, verbose=3))

        # Setting of node count properties
        coord = f[0].construct("axis=X")
        nc = coord.get_node_count()
        cfdm.write(f, tempfile)
        nc.set_property("long_name", "Node counts")
        cfdm.write(f, tempfile)
        nc.nc_set_variable("new_var_name_X")
        cfdm.write(f, tempfile)

        # Node count access
        c = g.construct("longitude").copy()
        self.assertTrue(c.has_node_count())
        n = c.del_node_count()
        self.assertFalse(c.has_node_count())
        self.assertIsNone(c.get_node_count(None))
        self.assertIsNone(c.del_node_count(None))
        c.set_node_count(n)
        self.assertTrue(c.has_node_count())
        self.assertTrue(c.get_node_count(None).equals(n, verbose=3))
        self.assertTrue(c.del_node_count(None).equals(n, verbose=3))
        self.assertFalse(c.has_node_count())

    def test_geometry_2(self):
        """Test nodes not tied to auxiliary coordinate variables."""
        f = cfdm.read(self.geometry_2_file, verbose=False)

        self.assertEqual(len(f), 2, "f = " + repr(f))

        for g in f:
            self.assertTrue(g.equals(g.copy(), verbose=3))
            self.assertEqual(len(g.auxiliary_coordinates()), 3)

        g = f[0]
        for axis in ("X", "Y", "Z"):
            coord = g.construct("axis=" + axis)
            self.assertTrue(coord.has_node_count(), "axis=" + axis)
            self.assertFalse(coord.has_part_node_count(), "axis=" + axis)
            self.assertFalse(coord.has_interior_ring(), "axis=" + axis)

        cfdm.write(f, tempfile, Conventions="CF-" + VN, verbose=False)

        f2 = cfdm.read(tempfile, verbose=False)

        self.assertEqual(len(f2), 2, "f2 = " + repr(f2))

        for a, b in zip(f, f2):
            self.assertTrue(a.equals(b, verbose=3))

        # Setting of node count properties
        coord = f[0].construct("axis=X")
        nc = coord.get_node_count()
        cfdm.write(f, tempfile)
        nc.set_property("long_name", "Node counts")
        cfdm.write(f, tempfile, verbose=False)
        nc.nc_set_variable("new_var_name")
        cfdm.write(f, tempfile, verbose=False)

    def test_geometry_3(self):
        """Test nodes in a file with no node count variable."""
        f = cfdm.read(self.geometry_3_file, verbose=False)

        self.assertEqual(len(f), 2, "f = " + repr(f))

        for g in f:
            self.assertTrue(g.equals(g.copy(), verbose=3))
            self.assertEqual(len(g.auxiliary_coordinates()), 3)

        g = f[0]
        for axis in ("X", "Y", "Z"):
            coord = g.construct("axis=" + axis)
            self.assertFalse(coord.has_node_count(), "axis=" + axis)
            self.assertFalse(coord.has_part_node_count(), "axis=" + axis)
            self.assertFalse(coord.has_interior_ring(), "axis=" + axis)

        cfdm.write(f, tempfile, Conventions="CF-" + VN, verbose=False)

        f2 = cfdm.read(tempfile, verbose=False)

        self.assertEqual(len(f2), 2, "f2 = " + repr(f2))

        for a, b in zip(f, f2):
            self.assertTrue(a.equals(b, verbose=3))

    def test_geometry_4(self):
        """Test nodes all not tied to auxiliary coordinate variables."""
        f = cfdm.read(self.geometry_4_file, verbose=False)

        self.assertEqual(len(f), 2, "f = " + repr(f))

        for g in f:
            self.assertTrue(g.equals(g.copy(), verbose=3))
            self.assertEqual(len(g.auxiliary_coordinates()), 3)

        for axis in ("X", "Y"):
            coord = g.construct("axis=" + axis)
            self.assertTrue(coord.has_node_count(), "axis=" + axis)
            self.assertFalse(coord.has_part_node_count(), "axis=" + axis)
            self.assertFalse(coord.has_interior_ring(), "axis=" + axis)

        cfdm.write(f, tempfile, Conventions="CF-" + VN, verbose=False)

        f2 = cfdm.read(tempfile, verbose=False)

        self.assertEqual(len(f2), 2, "f2 = " + repr(f2))

        for a, b in zip(f, f2):
            self.assertTrue(a.equals(b, verbose=3))

        # Setting of node count properties
        coord = f[0].construct("axis=X")
        nc = coord.get_node_count()
        cfdm.write(f, tempfile)
        nc.set_property("long_name", "Node counts")
        cfdm.write(f, tempfile, verbose=False)
        nc.nc_set_variable("new_var_name")
        cfdm.write(f, tempfile, verbose=False)

    def test_geometry_interior_ring(self):
        """Test the management of interior ring geometries."""
        for geometry_file in (
            self.geometry_interior_ring_file,
            self.geometry_interior_ring_file_2,
        ):
            f = cfdm.read(geometry_file, verbose=False)

            self.assertEqual(len(f), 2, "f = " + repr(f))

            for g in f:
                self.assertTrue(g.equals(g.copy(), verbose=3))
                self.assertEqual(len(g.auxiliary_coordinates()), 4)

            g = f[0]
            for axis in ("X", "Y"):
                coord = g.construct("axis=" + axis)
                self.assertTrue(coord.has_node_count(), "axis=" + axis)
                self.assertTrue(coord.has_part_node_count(), "axis=" + axis)
                self.assertTrue(coord.has_interior_ring(), "axis=" + axis)

            cfdm.write(f, tempfile, Conventions="CF-" + VN)

            f2 = cfdm.read(tempfile)

            self.assertEqual(len(f2), 2, "f2 = " + repr(f2))

            for a, b in zip(f, f2):
                self.assertTrue(a.equals(b, verbose=3))

            # Interior ring component
            c = g.construct("longitude")

            self.assertTrue(
                c.interior_ring.equals(
                    g.construct("longitude").get_interior_ring()
                )
            )
            self.assertEqual(c.interior_ring.data.ndim, c.data.ndim + 1)
            self.assertEqual(c.interior_ring.data.shape[0], c.data.shape[0])

            self.assertIsInstance(g.dump(display=False), str)

            d = c.insert_dimension(0)
            self.assertEqual(d.data.shape, (1,) + c.data.shape)
            self.assertEqual(
                d.interior_ring.data.shape, (1,) + c.interior_ring.data.shape
            )

            e = d.squeeze(0)
            self.assertEqual(e.data.shape, c.data.shape)
            self.assertEqual(
                e.interior_ring.data.shape, c.interior_ring.data.shape
            )

            t = d.transpose()
            self.assertEqual(
                t.data.shape,
                d.data.shape[::-1],
                (t.data.shape, c.data.shape[::-1]),
            )
            self.assertEqual(
                t.interior_ring.data.shape,
                (
                    d.interior_ring.data.shape[-2::-1]
                    + (d.interior_ring.data.shape[-1],)
                ),
            )

            # Subspacing
            g = g[1, ...]
            c = g.construct("longitude")
            self.assertEqual(c.interior_ring.data.shape[0], 1)
            self.assertEqual(c.interior_ring.data.ndim, c.data.ndim + 1)
            self.assertEqual(c.interior_ring.data.shape[0], c.data.shape[0])

            # Setting of node count properties
            coord = f[0].construct("axis=Y")
            nc = coord.get_node_count()
            nc.set_property("long_name", "Node counts")
            cfdm.write(f, tempfile)

            nc.nc_set_variable("new_var_name")
            cfdm.write(f, tempfile)

            # Setting of part node count properties
            coord = f[0].construct("axis=X")
            pnc = coord.get_part_node_count()
            pnc.set_property("long_name", "Part node counts")
            cfdm.write(f, tempfile)

            pnc.nc_set_variable("new_var_name")
            cfdm.write(f, tempfile)

            pnc.nc_set_dimension("new_dim_name")
            cfdm.write(f, tempfile)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print()
    unittest.main(verbosity=2)



import atexit
import datetime
import faulthandler
import os
import re
import tempfile
import unittest

import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

n_tmpfiles = 1
tmpfiles = [
    tempfile.mkstemp("_test_Field.nc", dir=os.getcwd())[1]
    for i in range(n_tmpfiles)
]
(tmpfile,) = tmpfiles


def _remove_tmpfiles():
    """Remove temporary files created during tests."""
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass


atexit.register(_remove_tmpfiles)


class FieldTest(unittest.TestCase):
    """Unit test for the Field class."""

    f0 = cfdm.example_field(0)
    f1 = cfdm.example_field(1)

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_Field__repr__str__dump_construct_type(self):
        """Test all means of Field inspection."""
        f = self.f1

        repr(f)
        str(f)
        self.assertIsInstance(f.dump(display=False), str)
        self.assertEqual(f.construct_type, "field")

        # Test when any construct which can have data in fact has no data.
        f = f.copy()
        for identity in [
            "time",  # a dimension coordinate
            "latitude",  # an auxiliary coordinate
            "measure:area",  # a cell measure
            "surface_altitude",  # a domain ancillary,
            "air_temperature standard_error",  # a field ancillary
        ]:
            c = f.construct(identity)  # get relevant construct, type as above
            c.del_data()
            self.assertFalse(c.has_data())
            str(f)
            repr(f)

    def test_Field__init__(self):
        """Test the Field constructor and source keyword."""
        cfdm.Field(source="qwerty")

    def test_Field__getitem__(self):
        """Test the access of field subspsaces from Field."""
        f = self.f1
        f = f.squeeze()

        d = f.data.array

        g = f[...]
        self.assertTrue((g.data.array == d).all())

        g = f[:, :]
        self.assertTrue((g.data.array == d).all())

        g = f[slice(None), :]
        self.assertTrue((g.data.array == d).all())

        g = f[slice(None), slice(None)]
        self.assertTrue((g.data.array == d).all())

        g = f[slice(None), ...]
        self.assertTrue((g.data.array == d).all())

        g = f[..., slice(None)]
        self.assertTrue((g.data.array == d).all())

        g = f[:, slice(0, f.data.shape[1], 1)]
        self.assertTrue((g.data.array == d).all())

        for indices, shape, multiple_list_indices in (
            [(slice(0, None, 1), slice(0, None)), (10, 9), False],
            [(slice(3, 7), slice(2, 5)), (4, 3), False],
            [(slice(6, 2, -1), slice(4, 1, -1)), (4, 3), False],
            # [(1, 3), ( 1, 1), False],
            # [(-2, -4), ( 1, 1), False],
            # [(-2, slice(1, 5)), ( 1, 4), False],
            # [(slice(5, 1, -2), 7), ( 2, 1), False],
            [([1, 4, 7], slice(1, 5)), (3, 4), False],
            [([1, 4, 7], slice(6, 8)), (3, 2), False],
            [(slice(6, 8), [1, 4, 7]), (2, 3), False],
            [([0, 3, 8], [1, 7, 8]), (3, 3), True],
            [([8, 3, 0], [8, 7, 1]), (3, 3), True],
        ):
            g = f[indices]

            if not multiple_list_indices:
                e = d[indices]
            else:
                e = d.copy()
                indices = list(indices)
                for axis, i in enumerate(indices):
                    if isinstance(i, list):
                        e = np.take(e, indices=i, axis=axis)
                        indices[axis] = slice(None)

                e = e[tuple(indices)]

            self.assertEqual(
                g.data.shape,
                e.data.shape,
                f"Bad shape for {indices}: {g.data.shape} != {e.data.shape}",
            )
            self.assertTrue(
                (g.data.array == e).all(),
                f"Bad values for {indices}: {g.data.array} != {e}",
            )

        # Check slicing of bounds
        g = f[..., 0:4]
        c = g.construct("grid_longitude")
        b = c.bounds
        self.assertEqual(c.data.shape, (4,))
        self.assertEqual(b.data.shape, (4, 2))

        # Indices result in a subspaced shape that has a size 0 axis
        with self.assertRaises(IndexError):
            f[..., [False] * f.shape[-1]]

    def test_Field_get_filenames(self):
        """Test Field.get_filenames."""
        cfdm.write(self.f0, tmpfile)
        f = cfdm.read(tmpfile)[0]

        abspath_tmpfile = os.path.abspath(tmpfile)
        self.assertEqual(f.get_filenames(), set([abspath_tmpfile]))

        f.persist(inplace=True, metadata=True)
        self.assertEqual(f.get_filenames(), set())

    def test_Field_apply_masking(self):
        """Test the `apply_masking` Field method."""
        f = self.f0.copy()

        for prop in (
            "missing_value",
            "_FillValue",
            "valid_min",
            "valid_max",
            "valid_range",
        ):
            f.del_property(prop, None)

        d = f.data.copy()
        g = f.copy()
        self.assertIsNone(f.apply_masking(inplace=True))
        self.assertTrue(f.equals(g, verbose=3))

        x = 0.11
        y = 0.1
        z = 0.2

        f.set_property("_FillValue", x)
        d = f.data.copy()

        g = f.apply_masking()
        e = d.apply_masking(fill_values=[x])
        self.assertTrue(e.equals(g.data, verbose=3))
        self.assertEqual(g.data.array.count(), g.data.size - 1)

        f.set_property("valid_range", [y, z])
        d = f.data.copy()
        g = f.apply_masking()
        e = d.apply_masking(fill_values=[x], valid_range=[y, z])
        self.assertTrue(e.equals(g.data, verbose=3))

        f.del_property("valid_range")
        f.set_property("valid_min", y)
        g = f.apply_masking()
        e = d.apply_masking(fill_values=[x], valid_min=y)
        self.assertTrue(e.equals(g.data, verbose=3))

        f.del_property("valid_min")
        f.set_property("valid_max", z)
        g = f.apply_masking()
        e = d.apply_masking(fill_values=[x], valid_max=z)
        self.assertTrue(e.equals(g.data, verbose=3))

        f.set_property("valid_min", y)
        g = f.apply_masking()
        e = d.apply_masking(fill_values=[x], valid_min=y, valid_max=z)
        self.assertTrue(e.equals(g.data, verbose=3))

    def test_Field_PROPERTIES(self):
        """Test the property access methods of Field."""
        f = self.f1.copy()
        for name, value in f.properties().items():
            self.assertTrue(f.has_property(name))
            f.get_property(name)
            f.del_property(name)
            self.assertIsNone(f.del_property(name, default=None))
            self.assertIsNone(f.get_property(name, default=None))
            self.assertFalse(f.has_property(name))
            f.set_property(name, value)

        d = f.clear_properties()
        self.assertIsInstance(d, dict)

        f.set_properties(d)
        f.set_properties(d, copy=False)

    def test_Field_set_get_del_has_data(self):
        """Test the data access and (un)setting methods of Field."""
        f = self.f1.copy()

        self.assertTrue(f.has_data())
        data = f.get_data()
        self.assertIsInstance(f.del_data(), cfdm.Data)
        self.assertIsNone(f.get_data(default=None))
        self.assertIsNone(f.del_data(default=None))
        self.assertFalse(f.has_data())
        self.assertIsNone(f.set_data(data, axes=None))
        self.assertIsNone(f.set_data(data, axes=None, copy=False))
        self.assertTrue(f.has_data())

        f = self.f1.copy()
        self.assertIsInstance(f.del_data_axes(), tuple)
        self.assertFalse(f.has_data_axes())
        self.assertIsNone(f.del_data_axes(default=None))

        f = self.f1.copy()
        for key in f.constructs.filter_by_data():
            self.assertTrue(f.has_data_axes(key))
            self.assertIsInstance(f.get_data_axes(key), tuple)
            self.assertIsInstance(f.del_data_axes(key), tuple)
            self.assertIsNone(f.del_data_axes(key, default=None))
            self.assertIsNone(f.get_data_axes(key, default=None))
            self.assertFalse(f.has_data_axes(key))

        # Test inplace
        f = self.f1.copy()
        d = f.del_data()
        g = f.set_data(d, inplace=False)
        self.assertIsInstance(g, cfdm.Field)
        self.assertFalse(f.has_data())
        self.assertTrue(g.has_data())
        self.assertTrue(g.data.equals(d))

    def test_Field_construct_item(self):
        """Test the `construct_item` Field method."""
        f = self.f1

        out = f.construct_item("key%domainaxis0")
        self.assertEqual(len(out), 2)
        self.assertEqual(out[0], "domainaxis0")
        self.assertIsInstance(out[1], cfdm.DomainAxis)

    def test_Field_CONSTRUCTS(self):
        """Test the construct access Field methods."""
        f = self.f1

        f.construct("latitude")
        self.assertIsNone(f.construct("NOT_latitude", default=None))
        self.assertIsNone(f.construct(re.compile("^l"), default=None))

        key = f.construct_key("latitude")
        f.get_construct(key)
        self.assertIsNone(f.get_construct("qwerty", default=None))

        constructs = f.auxiliary_coordinates()
        n = 3
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.AuxiliaryCoordinate)

        constructs = f.cell_measures()
        n = 1
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.CellMeasure)

        constructs = f.cell_methods()
        n = 2
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.CellMethod)

        # Check that no cell methods are returned when a bad
        # identifier is provided
        self.assertFalse(f.cell_methods("bad identifier"))

        constructs = f.coordinate_references()
        n = 2
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.CoordinateReference)

        constructs = f.dimension_coordinates()
        n = 4
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DimensionCoordinate)

        constructs = f.domain_ancillaries()
        n = 3
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DomainAncillary)

        constructs = f.field_ancillaries()
        n = 1
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.FieldAncillary)

        constructs = f.domain_topologies()
        n = 0
        self.assertEqual(len(constructs), n)

        constructs = f.cell_connectivities()
        n = 0
        self.assertEqual(len(constructs), n)

        # Domain axis key
        ckey = f.construct_key("grid_latitude")
        dakey = f.get_data_axes(ckey)[0]
        self.assertEqual(f.domain_axis_key("grid_latitude"), dakey)
        self.assertIsNone(f.domain_axis_key("XXXX_latitude", default=None))
        self.assertIsNone(
            f.domain_axis_key(re.compile("^grid_"), default=None)
        )

        # UGRID
        f = cfdm.example_field(8)

        constructs = f.domain_topologies()
        n = 1
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.DomainTopology)

        constructs = f.cell_connectivities()
        n = 1
        self.assertEqual(len(constructs), n)
        for key, value in constructs.items():
            self.assertIsInstance(value, cfdm.CellConnectivity)

    def test_Field_domain_axes(self):
        """Test the `domain_axes` Field method."""
        f = self.f1

        regex = re.compile("^atmos")

        self.assertEqual(len(f.domain_axes()), 4)
        self.assertEqual(len(f.domain_axes("grid_latitude", -1)), 2)
        self.assertEqual(len(f.domain_axes(regex)), 1)
        self.assertEqual(len(f.domain_axes(regex, "grid_latitude", -1)), 3)

    def test_Field_data_axes(self):
        """Test the data axes access and (un)setting Field methods."""
        f = self.f1.copy()

        ref = f.get_data_axes()

        self.assertEqual(f.get_data_axes(default=None), ref)

        self.assertEqual(f.del_data_axes(), ref)
        self.assertIsNone(f.del_data_axes(default=None))

        self.assertIsNone(f.set_data_axes(ref))
        self.assertEqual(f.get_data_axes(), ref)

    def test_Field_convert(self):
        """Test the convert Field method."""
        f = self.f1

        key = f.construct_key("grid_latitude")
        c = f.convert(key)

        self.assertEqual(c.data.ndim, 1)
        self.assertEqual(c.get_property("standard_name"), "grid_latitude")
        self.assertEqual(len(c.dimension_coordinates()), 1)
        self.assertEqual(len(c.auxiliary_coordinates()), 1)
        self.assertEqual(len(c.cell_measures()), 0)
        self.assertEqual(len(c.coordinate_references()), 1)
        self.assertEqual(len(c.domain_ancillaries()), 0)
        self.assertEqual(len(c.field_ancillaries()), 0)
        self.assertEqual(len(c.cell_methods()), 0)

        key = f.construct_key("latitude")
        c = f.convert(key)

        self.assertEqual(c.data.ndim, 2)
        self.assertEqual(c.get_property("standard_name"), "latitude")
        self.assertEqual(len(c.dimension_coordinates()), 2)
        self.assertEqual(len(c.auxiliary_coordinates()), 3)
        self.assertEqual(len(c.cell_measures()), 1)
        self.assertEqual(len(c.coordinate_references()), 1)
        self.assertEqual(len(c.domain_ancillaries()), 0)
        self.assertEqual(len(c.field_ancillaries()), 0)
        self.assertEqual(len(c.cell_methods()), 0)

        with self.assertRaises(ValueError):
            f.convert("qwerty")

        # Test some constructs which can never have data
        with self.assertRaises(ValueError):
            f.convert("cellmethod0")
        with self.assertRaises(ValueError):
            f.convert("domainaxis0")

    def test_Field_equals(self):
        """Test the equality-testing Field method."""
        f = self.f1

        self.assertTrue(f.equals(f, verbose=3))

        g = f.copy()
        self.assertTrue(f.equals(g, verbose=3))
        self.assertTrue(g.equals(f, verbose=3))

        g = f[...]
        self.assertTrue(f.equals(g, verbose=3))
        self.assertTrue(g.equals(f, verbose=3))

        g = g.squeeze()
        self.assertFalse(f.equals(g))

        h = f.copy()
        h.data[...] = h.data.array[...] + 1
        self.assertFalse(f.equals(h))

        # Symmetry
        f = cfdm.example_field(2)
        g = f.copy()
        self.assertTrue(f.equals(g))
        self.assertTrue(g.equals(f))

        g.del_construct("dimensioncoordinate0")
        self.assertFalse(f.equals(g))
        self.assertFalse(g.equals(f))

    def test_Field_del_construct(self):
        """Test the `del_construct` Field method."""
        f = self.f1.copy()

        self.assertIsInstance(
            f.del_construct("auxiliarycoordinate1"), cfdm.AuxiliaryCoordinate
        )

        with self.assertRaises(ValueError):
            f.del_construct("auxiliarycoordinate1")

        self.assertIsNone(
            f.del_construct("auxiliarycoordinate1", default=None)
        )

        self.assertIsInstance(
            f.del_construct("measure:area"), cfdm.CellMeasure
        )

    def test_Field_has_construct(self):
        """Test the `has_construct` Field method."""
        f = self.f1.copy()

        self.assertTrue(f.has_construct("latitude"))

        self.assertTrue(f.has_construct("auxiliarycoordinate1"))

        self.assertFalse(f.has_construct("QWERTY"))

        # Test edge case whereby constructs have Falsy values as key names:
        f.set_construct(cfdm.DomainAxis(0), key="")
        self.assertTrue(f.has_construct(""))

    def test_Field_squeeze(self):
        """Test Field.squeeze."""
        f = self.f1

        for axes in (None, "atmosphere_hybrid_height_coordinate"):
            g = f.squeeze(axes)
            self.assertEqual(g.data.shape, f.data.shape[1:])
            self.assertEqual(g.get_data_axes(), f.get_data_axes()[1:])

        self.assertIsNone(g.squeeze(inplace=True))

    def test_Field_transpose(self):
        """Test Field.transpose."""
        f = self.f1

        g = f.transpose()
        self.assertEqual(g.data.shape, f.data.shape[::-1])
        self.assertEqual(g.get_data_axes(), f.get_data_axes()[::-1])

        g = g.transpose(
            [
                "atmosphere_hybrid_height_coordinate",
                "grid_latitude",
                "grid_longitude",
            ]
        )
        self.assertEqual(g.shape, f.shape)
        self.assertEqual(g.get_data_axes(), f.get_data_axes())

    def test_Field_AAAinsert_dimension(self):
        """Test cfdm.Field.insert_dimension method."""
        f = self.f1
        g = f.copy()

        key = g.set_construct(cfdm.DomainAxis(1))
        h = g.insert_dimension(axis=key)
        self.assertEqual(h.data.ndim, f.data.ndim + 1)
        self.assertEqual(h.get_data_axes()[1:], f.get_data_axes())

        key = g.set_construct(cfdm.DomainAxis(1))
        h = g.insert_dimension(position=g.data.ndim, axis=key)
        self.assertEqual(h.data.ndim, f.data.ndim + 1)
        self.assertEqual(h.get_data_axes()[:-1], f.get_data_axes())

        self.assertEqual(g.cell_measure().ndim, 2)
        h = g.insert_dimension(None, constructs=True)
        self.assertEqual(h.cell_measure().ndim, 3)

        f = f.squeeze()
        array = f.array
        for i in tuple(range(f.ndim + 1)) + tuple(range(-1, -f.ndim - 2, -1)):
            self.assertEqual(
                f.insert_dimension(None, i).shape,
                np.expand_dims(array, i).shape,
            )

    def test_Field_compress_uncompress(self):
        """Test the compress and uncompress Field methods."""
        contiguous = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "DSG_timeSeries_contiguous.nc",
        )
        indexed = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "DSG_timeSeries_indexed.nc",
        )
        indexed_contiguous = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "DSG_timeSeriesProfile_indexed_contiguous.nc",
        )

        files = (contiguous, indexed, indexed_contiguous)
        methods = ("contiguous", "indexed", "indexed_contiguous")

        for filename, method in zip(files, methods):
            message = "method=" + method
            for f in cfdm.read(filename):
                self.assertTrue(bool(f.data.get_compression_type()), message)

                u = f.uncompress()
                self.assertFalse(bool(u.data.get_compression_type()), message)
                self.assertTrue(f.equals(u, verbose=3), message)

                for method1 in methods:
                    message += ", method1=" + method1
                    if method1 == "indexed_contiguous":
                        if f.data.ndim != 3:
                            continue
                    elif f.data.ndim != 2:
                        continue

                    c = u.compress(method1)

                    self.assertTrue(
                        bool(c.data.get_compression_type()), message
                    )

                    self.assertTrue(u.equals(c, verbose=3), message)
                    self.assertTrue(f.equals(c, verbose=3), message)

                    c = f.compress(method1)
                    self.assertTrue(
                        bool(c.data.get_compression_type()), message
                    )

                    self.assertTrue(u.equals(c, verbose=3), message)
                    self.assertTrue(f.equals(c, verbose=3), message)

                    cfdm.write(c, tmpfile)
                    c = cfdm.read(tmpfile)[0]

                    self.assertTrue(
                        bool(c.data.get_compression_type()), message
                    )
                    self.assertTrue(f.equals(c, verbose=3), message)

    def test_Field_creation_commands(self):
        """Test the Field.creation_commands."""
        for i in range(12):
            f = cfdm.example_field(i)

        f = self.f1

        for rd in (False, True):
            f.creation_commands(representative_data=rd)

        for indent in (0, 4):
            f.creation_commands(indent=indent)

        for s in (False, True):
            f.creation_commands(string=s)

        for ns in ("cfdm", ""):
            f.creation_commands(namespace=ns)

    def test_Field_has_geometry(self):
        """Test Field.has_geometry."""
        f = self.f1
        self.assertFalse(f.has_geometry())

        f = cfdm.example_field(6)
        self.assertTrue(f.has_geometry())

    def test_Field_climatological_time_axes(self):
        """Test the `climatological_time_axes` method of Field."""
        f = self.f0.copy()
        self.assertEqual(f.climatological_time_axes(), set())

        f.set_construct(
            cfdm.CellMethod("domainaxis2", "mean", {"within": "years"})
        )
        f.set_construct(
            cfdm.CellMethod("domainaxis2", "mean", {"over": "years"})
        )

        cta = f.climatological_time_axes()
        self.assertEqual(cta, set(["domainaxis2"]))

        d = f.get_domain()
        self.assertEqual(d.climatological_time_axes(), cta)

    def test_Field_bounds(self):
        """Test that Field instances do not have cell bounds."""
        f = self.f0
        self.assertFalse(f.has_bounds())

    def test_Field_auxiliary_coordinate(self):
        """Test Field.auxiliary_coordinate."""
        f = self.f1

        for identity in ("auxiliarycoordinate1", "latitude"):
            key, c = f.construct(identity, item=True)
            self.assertTrue(f.auxiliary_coordinate(identity).equals(c))
            self.assertEqual(f.auxiliary_coordinate(identity, key=True), key)

        with self.assertRaises(ValueError):
            f.auxiliary_coordinate("long_name:qwerty")

    def test_Field_coordinate(self):
        """Test Field.coordinate."""
        f = self.f1

        for identity in (
            "latitude",
            "grid_longitude",
            "auxiliarycoordinate1",
            "dimensioncoordinate1",
        ):
            key, c = f.construct(identity, item=True)

        with self.assertRaises(ValueError):
            f.coordinate("long_name:qweRty")

    def test_Field_coordinate_reference(self):
        """Test Field.coordinate_reference."""
        f = self.f1

        for identity in (
            "coordinatereference1",
            "key%coordinatereference0",
            "standard_name:atmosphere_hybrid_height_coordinate",
            "grid_mapping_name:rotated_latitude_longitude",
        ):
            key, c = f.construct(identity, item=True)
            self.assertTrue(f.coordinate_reference(identity).equals(c))
            self.assertEqual(f.coordinate_reference(identity, key=True), key)

        with self.assertRaises(ValueError):
            f.coordinate_reference("qwerty")

    def test_Field_dimension_coordinate(self):
        """Test Field.dimension_coordinate."""
        f = self.f1

        for identity in ("grid_latitude", "dimensioncoordinate1"):
            if identity == "X":
                key, c = f.construct("grid_longitude", item=True)
            else:
                key, c = f.construct(identity, item=True)

            self.assertTrue(f.dimension_coordinate(identity).equals(c))
            self.assertEqual(f.dimension_coordinate(identity, key=True), key)

            k, v = f.dimension_coordinate(identity, item=True)
            self.assertEqual(k, key)
            self.assertTrue(v.equals(c))

        self.assertIsNone(
            f.dimension_coordinate("long_name=qwerty:asd", default=None)
        )
        self.assertEqual(
            len(f.dimension_coordinates("long_name=qwerty:asd")), 0
        )

        with self.assertRaises(ValueError):
            f.dimension_coordinate("long_name:qwerty")

    def test_Field_cell_measure(self):
        """Test Field.cell_measure."""
        f = self.f1

        for identity in ("measure:area", "cellmeasure0"):
            key, c = f.construct(identity, item=True)

            self.assertTrue(f.cell_measure(identity).equals(c))
            self.assertEqual(f.cell_measure(identity, key=True), key)

            self.assertTrue(f.cell_measure(identity).equals(c))
            self.assertEqual(f.cell_measure(identity, key=True), key)

        self.assertEqual(len(f.cell_measures()), 1)
        self.assertEqual(len(f.cell_measures("measure:area")), 1)
        self.assertEqual(len(f.cell_measures(*["measure:area"])), 1)

        self.assertIsNone(f.cell_measure("long_name=qwerty:asd", default=None))
        self.assertEqual(len(f.cell_measures("long_name=qwerty:asd")), 0)

        with self.assertRaises(ValueError):
            f.cell_measure("long_name:qwerty")

    def test_Field_cell_method(self):
        """Test Field.cell_method."""
        f = self.f1

        for identity in ("method:mean", "cellmethod0"):
            key, c = f.construct(identity, item=True)
            self.assertTrue(f.cell_method(identity).equals(c))
            self.assertEqual(f.cell_method(identity, key=True), key)

    def test_Field_domain_ancillary(self):
        """Test Field.domain_ancillary."""
        f = self.f1

        for identity in ("surface_altitude", "domainancillary0"):
            key, c = f.construct(identity, item=True)
            self.assertTrue(f.domain_ancillary(identity).equals(c))
            self.assertEqual(f.domain_ancillary(identity, key=True), key)

        with self.assertRaises(ValueError):
            f.domain_ancillary("long_name:qwerty")

    def test_Field_field_ancillary(self):
        """Test Field.field_ancillary."""
        f = self.f1

        for identity in ("air_temperature standard_error", "fieldancillary0"):
            key, c = f.construct_item(identity)
            self.assertTrue(f.field_ancillary(identity).equals(c))
            self.assertEqual(f.field_ancillary(identity, key=True), key)

        with self.assertRaises(ValueError):
            f.field_ancillary("long_name:qwerty")

    def test_Field_domain_axis(self):
        """Test Field.domain_axis."""
        f = self.f1

        f.domain_axis(1)
        f.domain_axis("domainaxis2")

        with self.assertRaises(ValueError):
            f.domain_axis(99)

        with self.assertRaises(ValueError):
            f.domain_axis("qwerty")

    def test_Field_indices(self):
        """Test Field.indices."""
        f = self.f0

        g = f[f.indices(longitude=112.5)]
        self.assertEqual(g.shape, (5, 1))
        x = g.dimension_coordinate("longitude").data.array
        self.assertTrue((x == 112.5).all())

        g = f[f.indices(longitude=112.5, latitude=[-45, 75])]
        self.assertEqual(g.shape, (2, 1))
        x = g.dimension_coordinate("longitude").data.array
        y = g.dimension_coordinate("latitude").data.array
        self.assertTrue((x == 112.5).all())
        self.assertTrue((y == [-45, 75]).all())

        g = f[f.indices(time=31)]
        self.assertTrue(g.equals(f))

        g = f[f.indices(time=np.array([31, 9999]))]
        self.assertTrue(g.equals(f))

        with self.assertRaises(ValueError):
            f.indices(bad_name=23)

        with self.assertRaises(ValueError):
            f.indices(longitude=-999)

        with self.assertRaises(ValueError):
            f.indices(longitude="bad_value_type")

        # Test for same axis specified twice
        key = f.construct("longitude", key=True)
        with self.assertRaises(ValueError):
            f.indices(**{"longitude": 112.5, key: 22.5})

    def test_Field_get_original_filenames(self):
        """Test Field.orignal_filenames."""
        f = self.f0
        f._original_filenames(define=["file1.nc", "file2.nc"])
        x = f.coordinate("longitude")
        x._original_filenames(define=["file1.nc", "file3.nc"])
        b = x.bounds
        b._original_filenames(define=["file1.nc", "file4.nc"])

        self.assertEqual(
            f.get_original_filenames(),
            set(
                (
                    cfdm.abspath("file1.nc"),
                    cfdm.abspath("file2.nc"),
                    cfdm.abspath("file3.nc"),
                    cfdm.abspath("file4.nc"),
                )
            ),
        )

        self.assertEqual(
            f.get_original_filenames(), f.copy().get_original_filenames()
        )

    def test_Field_del_properties(self):
        """Test the del_properties method of Field."""
        f = cfdm.Field()
        f.set_properties({"project": "CMIP7", "comment": "model"})
        properties = f.properties()

        removed_properties = f.del_properties("project")
        self.assertEqual(removed_properties, {"project": "CMIP7"})
        self.assertEqual(f.properties(), {"comment": "model"})
        f.set_properties(removed_properties)
        self.assertEqual(f.properties(), properties)
        self.assertEqual(f.del_properties("foo"), {})
        self.assertEqual(f.properties(), properties)

        removed_properties = f.del_properties(["project", "comment"])
        self.assertEqual(removed_properties, properties)
        self.assertEqual(f.properties(), {})

    def test_Field_hdf5_chunksizes(self):
        """Test the HDF5 chunk size methods of a Field."""
        f = self.f0.copy()

        f.nc_set_hdf5_chunksizes({"latitude": 1})
        self.assertEqual(f.nc_hdf5_chunksizes(), (1, 8))
        f.nc_set_hdf5_chunksizes({"longitude": 7})
        self.assertEqual(f.nc_hdf5_chunksizes(), (1, 7))
        f.nc_set_hdf5_chunksizes({"latitude": 4, "longitude": 2})
        self.assertEqual(f.nc_hdf5_chunksizes(), (4, 2))
        f.nc_set_hdf5_chunksizes([1, 7])
        self.assertEqual(f.nc_hdf5_chunksizes(), (1, 7))
        f.nc_set_hdf5_chunksizes("contiguous")
        self.assertEqual(f.nc_hdf5_chunksizes(), "contiguous")
        f.nc_set_hdf5_chunksizes(None)
        self.assertIsNone(f.nc_hdf5_chunksizes())

        for c in (64, "64 B"):
            f.nc_set_hdf5_chunksizes(c)
            self.assertEqual(f.nc_hdf5_chunksizes(), 64)

        f.nc_set_hdf5_chunksizes({"latitude": 999}, clip=True)
        self.assertEqual(f.nc_hdf5_chunksizes(), (5, 8))

        f.nc_set_hdf5_chunksizes({"latitude": 4, "time": 1})
        self.assertEqual(f.nc_hdf5_chunksizes(), (4, 8))
        for coord in ("time", "latitude", "longitude"):
            self.assertIsNone(
                f.dimension_coordinate(coord).nc_hdf5_chunksizes()
            )

        # todict
        f.nc_set_hdf5_chunksizes([3, 4])
        self.assertEqual(
            f.nc_hdf5_chunksizes(todict=True),
            {"time": 1, "latitude": 3, "longitude": 4},
        )

        for chunksizes in (None, "contiguous", 1024):
            f.nc_set_hdf5_chunksizes(chunksizes)
            with self.assertRaises(ValueError):
                f.nc_hdf5_chunksizes(todict=True)

        # ignore keyword
        f.nc_clear_hdf5_chunksizes()
        f.nc_set_hdf5_chunksizes({"latitude": 4, "BAD_axis": 99}, ignore=True)
        self.assertEqual(f.nc_hdf5_chunksizes(), (4, 8))
        with self.assertRaises(ValueError):
            f.nc_set_hdf5_chunksizes({"latitude": 4, "BAD_axis": 99})

        # filter_kwargs keyword
        f.nc_set_hdf5_chunksizes({"latitude": 4}, filter_by_naxes=(1,))
        self.assertEqual(f.nc_hdf5_chunksizes(), (4, 8))

        # constructs keyword
        f.nc_set_hdf5_chunksizes({"latitude": 4, "time": 1}, constructs=True)
        self.assertEqual(f.nc_hdf5_chunksizes(), (4, 8))

        f.nc_set_hdf5_chunksizes({"latitude": 4, "time": 1}, constructs=True)
        self.assertEqual(
            f.dimension_coordinate("time").nc_hdf5_chunksizes(), (1,)
        )
        self.assertEqual(
            f.dimension_coordinate("latitude").nc_hdf5_chunksizes(), (4,)
        )
        self.assertEqual(
            f.dimension_coordinate("longitude").nc_hdf5_chunksizes(), (8,)
        )

        f.nc_set_hdf5_chunksizes(
            "contiguous", constructs={"filter_by_axis": ("longitude",)}
        )
        self.assertEqual(f.nc_hdf5_chunksizes(), "contiguous")
        self.assertEqual(
            f.dimension_coordinate("time").nc_hdf5_chunksizes(), (1,)
        )
        self.assertEqual(
            f.dimension_coordinate("latitude").nc_hdf5_chunksizes(), (4,)
        )
        self.assertEqual(
            f.dimension_coordinate("longitude").nc_hdf5_chunksizes(),
            "contiguous",
        )

        # clear
        f.nc_set_hdf5_chunksizes("contiguous", constructs=True)
        self.assertEqual(f.nc_clear_hdf5_chunksizes(), "contiguous")
        self.assertIsNone(
            f.nc_clear_hdf5_chunksizes(
                constructs={"filter_by_identity": ("longitude",)}
            )
        )
        self.assertEqual(
            f.dimension_coordinate("latitude").nc_hdf5_chunksizes(),
            "contiguous",
        )
        self.assertIsNone(
            f.dimension_coordinate("longitude").nc_hdf5_chunksizes()
        )

        f.nc_clear_hdf5_chunksizes(constructs={})
        self.assertIsNone(
            f.dimension_coordinate("latitude").nc_hdf5_chunksizes()
        )

    def test_Field_concatenate(self):
        """Test Field.concatenate."""
        f = self.f1.copy()

        g = cfdm.Field.concatenate([f.copy()], axis=0)
        self.assertEqual(g.shape, (1, 10, 9))

        x = [f.copy() for i in range(8)]

        g = cfdm.Field.concatenate(x, axis=0)
        self.assertEqual(g.shape, (8, 10, 9))

        key = x[3].construct_key("latitude")
        x[3].del_construct(key)
        g = cfdm.Field.concatenate(x, axis=0)
        self.assertEqual(g.shape, (8, 10, 9))

        with self.assertRaises(Exception):
            g = cfdm.Field.concatenate([], axis=0)

    def test_Field_persist(self):
        """Test Field.persist."""
        f = self.f0.copy()
        cfdm.write(f, tmpfile)
        f = cfdm.read(tmpfile)[0]

        for d in (f.data.todict(), f.coordinate("longitude").data.todict()):
            on_disk = False
            for v in d.values():
                if isinstance(v, cfdm.data.abstract.FileArray):
                    on_disk = True

            self.assertTrue(on_disk)

        g = f.persist()
        d = g.data.todict()
        in_memory = False
        for v in d.values():
            if isinstance(v, np.ndarray):
                in_memory = True

        self.assertTrue(in_memory)

        d = g.coordinate("longitude").data.todict()
        on_disk = False
        for v in d.values():
            if isinstance(v, cfdm.data.abstract.FileArray):
                on_disk = True

        self.assertTrue(on_disk)

        # In-place and metdata
        f = cfdm.read(tmpfile)[0]
        self.assertIsNone(f.persist(metadata=True, inplace=True))
        for d in (
            f.data.todict(),
            f.coordinate("longitude").data.todict(),
        ):
            in_memory = False
            for v in d.values():
                if isinstance(v, np.ndarray):
                    in_memory = True

            self.assertTrue(in_memory)

    def test_Field_unsqueeze(self):
        """Test Field.unsqueeze."""
        f = self.f0.copy()
        self.assertEqual(f.shape, (5, 8))
        g = f.unsqueeze()
        self.assertEqual(g.shape, (1, 5, 8))
        self.assertIsNone(g.unsqueeze(inplace=True))
        self.assertEqual(g.shape, (1, 5, 8))


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



import datetime
import faulthandler
import os
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm


class NodeCountPropertiesTest(unittest.TestCase):
    """Unit test for the NodeCountProperties class."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        # cfdm.log_level('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

        self.geometry_interior_ring_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "geometry_interior_ring.nc",
        )
        self.geometry_interior_ring_file_2 = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "geometry_interior_ring_2.nc",
        )

    def test_NodeCountProperties__repr__str__dump(self):
        """Test all means of NodeCountProperties inspection."""
        f = cfdm.read(self.geometry_interior_ring_file)[0]

        coord = f.construct("axis=X")
        self.assertTrue(coord.has_node_count())

        n = coord.get_node_count()

        _ = repr(n)
        _ = str(n)
        self.assertIsInstance(n.dump(display=False), str)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print()
    unittest.main(verbosity=2)



import datetime
import faulthandler
import os
import unittest

import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

verbose = False
warnings = False


class create_fieldTest(unittest.TestCase):
    """Test ab initio creation of field constructs in memory."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or calls (those
        # without a 'verbose' option to do the same) e.g. to debug them, wrap
        # them (for methods, start-to-end internally) as follows:
        # cfdm.log_level('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

        self.filename = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "test_file.nc"
        )

        try:
            os.remove(self.filename)
        except Exception:
            pass

    def test_create_field(self):
        """Test ab initio creation of a first variation of field."""
        # Dimension coordinates
        dim1 = cfdm.DimensionCoordinate(data=cfdm.Data(np.arange(10.0)))
        dim1.set_property("standard_name", "grid_latitude")
        dim1.set_property("units", "degrees")

        data = np.arange(9.0) + 20
        data[-1] = 34
        dim0 = cfdm.DimensionCoordinate(data=cfdm.Data(data))
        dim0.set_property("standard_name", "grid_longitude")
        dim0.set_property("units", "degrees")
        dim0.nc_set_variable("x")

        array = dim0.data.array

        array = np.array([array - 0.5, array + 0.5]).transpose((1, 0))
        array[-2, 1] = 30
        array[-1, :] = [30, 36]
        dim0.set_bounds(cfdm.Bounds(data=cfdm.Data(array)))

        dim2 = cfdm.DimensionCoordinate(
            data=cfdm.Data([1.5]),
            bounds=cfdm.Bounds(data=cfdm.Data([[1, 2.0]])),
        )
        dim2.set_property(
            "standard_name", "atmosphere_hybrid_height_coordinate"
        )
        dim2.set_property("computed_standard_name", "altitude")

        # Auxiliary coordinates
        ak = cfdm.DomainAncillary(data=cfdm.Data([10.0]))
        ak.set_property("units", "m")
        ak.id = "atmosphere_hybrid_height_coordinate_ak"
        ak.set_bounds(cfdm.Bounds(data=cfdm.Data([[5, 15.0]])))

        bk = cfdm.DomainAncillary(data=cfdm.Data([20.0]))
        bk.id = "atmosphere_hybrid_height_coordinate_bk"
        bk.set_bounds(cfdm.Bounds(data=cfdm.Data([[14, 26.0]])))

        aux2 = cfdm.AuxiliaryCoordinate(
            data=cfdm.Data(np.arange(-45, 45, dtype="int32").reshape(10, 9))
        )
        aux2.set_property("units", "degree_N")
        aux2.set_property("standard_name", "latitude")

        aux3 = cfdm.AuxiliaryCoordinate(
            data=cfdm.Data(np.arange(60, 150, dtype="int32").reshape(9, 10))
        )
        aux3.set_property("standard_name", "longitude")
        aux3.set_property("units", "degreeE")

        array = np.ma.array(
            [
                "alpha",
                "beta",
                "gamma",
                "delta",
                "epsilon",
                "zeta",
                "eta",
                "theta",
                "iota",
                "kappa",
            ],
        )
        array[0] = np.ma.masked
        aux4 = cfdm.AuxiliaryCoordinate(data=cfdm.Data(array))
        aux4.set_property("long_name", "greek_letters")

        # Cell measures
        msr0 = cfdm.CellMeasure(
            data=cfdm.Data(1 + np.arange(90.0).reshape(9, 10) * 1234)
        )
        msr0.set_measure("area")
        msr0.set_property("units", "km2")
        msr0.nc_set_variable("areacella")

        # Data
        data = cfdm.Data(np.arange(90.0).reshape(10, 9))

        properties = {"units": "m s-1"}

        f = cfdm.Field(properties=properties)
        f.set_property("standard_name", "eastward_wind")

        da = cfdm.DomainAxis(9)
        da.nc_set_dimension("grid_longitude")
        axisX = f.set_construct(da)
        axisY = f.set_construct(cfdm.DomainAxis(10))
        axisZ = f.set_construct(cfdm.DomainAxis(1))

        f.set_data(data, axes=[axisY, axisX])

        x = f.set_construct(dim0, axes=axisX)
        y = f.set_construct(dim1, axes=axisY)
        z = f.set_construct(dim2, axes=[axisZ])

        lat = f.set_construct(aux2, axes=[axisY, axisX])
        lon = f.set_construct(aux3, axes=[axisX, axisY])
        f.set_construct(aux4, axes=[axisY])

        ak = f.set_construct(ak, axes=axisZ)
        bk = f.set_construct(bk, axes=[axisZ])

        # Coordinate references
        coordinate_conversion = cfdm.CoordinateConversion(
            parameters={
                "grid_mapping_name": "rotated_latitude_longitude",
                "grid_north_pole_latitude": 38.0,
                "grid_north_pole_longitude": 190.0,
            }
        )

        datum = cfdm.Datum(parameters={"earth_radius": 6371007})

        ref0 = cfdm.CoordinateReference(
            coordinate_conversion=coordinate_conversion,
            datum=datum,
            coordinates=[x, y, lat, lon],
        )

        f.set_construct(msr0, axes=[axisX, axisY])

        f.set_construct(ref0)

        orog = cfdm.DomainAncillary(data=f.get_data())
        orog.set_property("standard_name", "surface_altitude")
        orog.set_property("units", "m")
        orog = f.set_construct(orog, axes=[axisY, axisX])

        coordinate_conversion = cfdm.CoordinateConversion(
            parameters={
                "standard_name": "atmosphere_hybrid_height_coordinate",
                "computed_standard_name": "altitude",
            },
            domain_ancillaries={"orog": orog, "a": ak, "b": bk},
        )

        ref1 = cfdm.CoordinateReference(
            coordinates=[z],
            datum=datum,
            coordinate_conversion=coordinate_conversion,
        )

        ref1 = f.set_construct(ref1)

        # Field ancillary variables
        g = f.copy()
        anc = cfdm.FieldAncillary(data=g.get_data())
        anc.set_property("standard_name", "ancillaryA")
        f.set_construct(anc, axes=[axisY, axisX])

        g = f[0]
        g = g.squeeze()
        anc = cfdm.FieldAncillary(data=g.get_data())
        anc.set_property("long_name", "ancillaryB")
        f.set_construct(anc, axes=axisX)

        g = f[..., 0]
        g = g.squeeze()
        anc = cfdm.FieldAncillary(data=g.get_data())
        anc.set_property("foo", "bar")
        f.set_construct(anc, axes=[axisY])

        f.set_property("flag_values", np.array([1, 2, 4], "int32"))
        f.set_property("flag_meanings", "a bb ccc")
        f.set_property("flag_masks", [2, 1, 0])

        cm0 = cfdm.CellMethod(
            axes=axisX,
            method="mean",
            qualifiers={"interval": [cfdm.Data(1, "day")], "comment": "ok"},
        )

        cm1 = cfdm.CellMethod(
            axes=[axisY], method="maximum", qualifiers={"where": "sea"}
        )

        f.set_construct(cm0)
        f.set_construct(cm1)

        self.assertTrue(
            f.equals(f, verbose=verbose), "Field f not equal to itself"
        )

        self.assertTrue(
            f.equals(f.copy(), verbose=verbose),
            "Field f not equal to a copy of itself",
        )

        cfdm.write(f, self.filename, verbose=verbose)

        g = cfdm.read(self.filename, verbose=1)

        array = (
            g[0]
            .constructs.filter_by_identity("long_name=greek_letters")
            .value()
            .data.array
        )
        self.assertEqual(array[1], "beta")

        self.assertEqual(len(g), 1)

        g = g[0].squeeze()

        self.assertEqual(
            sorted(f.constructs),
            sorted(g.constructs),
            f"\n\nf (created in memory)"
            f"\n{f.constructs}"
            f"\n\n{f.constructs.items()}"
            f"\n\ng (read from disk)"
            f"\n{g.constructs}"
            f"\n\n{g.constructs.items()}",
        )

        self.assertTrue(
            f.equals(f, verbose=verbose),
            "Field f not equal to itself after having been written to disk",
        )

        self.assertTrue(
            f.equals(f.copy(), verbose=verbose),
            "Field f not equal to a copy of itself after having been "
            "written to disk",
        )

        self.assertTrue(
            g.equals(g.copy(), verbose=verbose),
            "Field g not equal to a copy of itself",
        )

        self.assertTrue(
            g.equals(f, verbose=verbose),
            "Field (f) not equal to itself read back in (g)",
        )

        x = g.dump(display=False)
        x = f.dump(display=False)

        g = cfdm.read(
            self.filename,
            verbose=verbose,
            extra="domain_ancillary",
            warnings=warnings,
        )


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



from numbers import Integral

import numpy as np
from dask.array.slicing import normalize_index
from dask.base import is_dask_collection

from ...functions import indices_shape, parse_indices


class IndexMixin:
    """Mixin class for lazy indexing of a data array.

    A data for a subspace is retrieved by casting the object as a
    `numpy` array. See `__getitem__` for more details.

    **Examples**

    >>> a = {{package}}.{{class}}(...)
    >>> a.shape
    (6, 5)
    >>> print(np.asanyarray(a)
    [[ 0  1  2  3  4])
     [ 5  6  7  8  9]
     [10 11 12 13 14]
     [15 16 17 18 19]
     [20 21 22 23 24]
     [25 26 27 28 29]]
    >>> a = a[::2, [1, 2, 4]]
    >>> a = a[[True, False, True], :]
    >>> a.shape
    (2, 3)
    >>> print(np.asanyarray(a))
    [[ 1,  2,  4],
     [21, 22, 24]]

    .. versionadded:: (cfdm) 1.11.2.0

    """

    def __array__(self, dtype=None, copy=None):
        """Convert the `{{class}}` into a `numpy` array.

        .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            dtype: optional
                Typecode or data-type to which the array is cast.

            copy: `None` or `bool`
                Included to match the v2 `numpy.ndarray.__array__`
                API, but ignored. The return numpy array is always
                independent.

                .. versionadded:: (cfdm) 1.12.0.0

        :Returns:

            `numpy.ndarray`
                An independent `numpy` array of the subspace of the
                data defined by the `indices` attribute.

        """
        array = self._get_array()
        if dtype is None:
            return array

        return array.astype(dtype, copy=False)

    def __getitem__(self, index):
        """Returns a subspace of the data as a new `{{class}}`.

        x.__getitem__(indices) <==> x[indices]

        Subspaces created by indexing are lazy and are not applied
        until the `{{class}}` object is converted to a `numpy` array,
        by which time all lazily-defined subspaces will have been
        converted to a single combined index which defines only the
        actual elements that need to be retrieved from the original
        data.

        The combined index is orthogonal, meaning that the index for
        each dimension is to be applied independently, regardless of
        how that index was defined. For instance, the indices ``[[0,
        1], [1, 3], 0]`` and ``[:2, 1::2, 0]`` will give identical
        results.

        For example, if the original data has shape ``(12, 145, 192)``
        and consecutive subspaces of ``[::2, [1, 3, 4], 96:]`` and
        ``[[0, 5], [True, False, True], 0]`` are applied, then only
        the elements defined by the combined index``[[0, 10], [1, 4],
        96]`` will be retrieved from the data when `__array__` is
        called.

        .. versionadded:: (cfdm) 1.11.2.0

        .. seealso:: `index`, `original_shape`, `__array__`,
                     `__getitem__`

        :Returns:

            `{{class}}`
                The subspaced data.

        """
        shape0 = self.shape
        index0 = self.index(conform=False)
        reference_shape = list(self.reference_shape)

        index1 = parse_indices(shape0, index, keepdims=False, newaxis=True)

        new = self.copy()
        new_indices = []
        new_shape = []

        newaxis = np.newaxis
        if len(index1) > len(index0):
            # Take new axes out of 'index1' for now. We'll put them
            # back later.
            none_positions = [
                i for i, ind1 in enumerate(index1) if ind1 is newaxis
            ]
            index1 = [ind1 for ind1 in index1 if ind1 is not newaxis]
        else:
            none_positions = []

        i = 0
        j = 0
        for ind0, reference_size in zip(index0, reference_shape[:]):
            if isinstance(ind0, Integral):
                # The previous call to __getitem__ resulted in a
                # dimension being removed (i.e. 'ind0' is
                # integer-valued). Therefore 'index1' must have fewer
                # elements than 'index0', so we need to "carry
                # forward" the integer-valued index so that it is
                # available at evaluation time.
                new_indices.append(ind0)
                continue

            ind1 = index1[i]
            size0 = shape0[i]

            i += 1
            if ind0 is newaxis:
                if isinstance(ind1, Integral):
                    # A previously introduced new axis is being
                    # removed by an integer index
                    if ind1 not in (0, -1):
                        raise IndexError(
                            f"index {ind1} is out of bounds for axis {i - 1} "
                            "with size 1"
                        )

                    reference_shape.pop(i - 1 - j)
                    j += 1
                else:
                    new_indices.append(ind0)

                continue

            # If this dimension is not subspaced by the new index then
            # we don't need to update the old index.
            if isinstance(ind1, slice) and ind1 == slice(None):
                new_indices.append(ind0)
                continue

            # Still here? Then we have to work out the index of the
            #             full array that is equivalent to applying
            #             'ind0' followed by 'ind1'.
            if is_dask_collection(ind1):
                # Note: This will never occur when this __getitem__ is
                #       being called from within a Dask graph, because
                #       any lazy indices will have already been
                #       computed as part of the whole graph execution;
                #       i.e. we don't have to worry about a
                #       compute-within-a-compute situation. (If this
                #       were not the case then we could add
                #       `scheduler="synchronous"` to the compute
                #       call.)
                ind1 = ind1.compute()

            if isinstance(ind0, slice):
                if isinstance(ind1, slice):
                    # ind0: slice
                    # ind1: slice
                    start, stop, step = ind0.indices(reference_size)
                    start1, stop1, step1 = ind1.indices(size0)
                    size1, mod1 = divmod(stop1 - start1, step1)

                    if mod1 != 0:
                        size1 += 1

                    start += start1 * step
                    step *= step1
                    stop = start + (size1 - 1) * step

                    if step > 0:
                        stop += 1
                    else:
                        stop -= 1

                    if stop < 0:
                        stop = None

                    new_index = slice(start, stop, step)
                else:
                    # ind0: slice
                    # ind1: int, or array of int/bool
                    new_index = np.arange(*ind0.indices(reference_size))[ind1]
            else:
                # ind0: array of int. If we made it to here then it
                #                     can't be anything else. This is
                #                     because we've dealt with ind0
                #                     being a slice or an int, the
                #                     very first ind0 is always
                #                     slice(None), and a previous ind1
                #                     that was an array of bool will
                #                     have resulted in this ind0 being
                #                     an array of int.
                #
                # ind1: anything
                new_index = np.asanyarray(ind0)[ind1]

            new_indices.append(new_index)

        if none_positions:
            for i in none_positions:
                new_indices.insert(i, newaxis)
                reference_shape.insert(i, 1)

        new._custom["index"] = tuple(new_indices)
        new._custom["reference_shape"] = tuple(reference_shape)

        # Find the shape defined by the new index
        new_shape = indices_shape(new_indices, reference_shape, keepdims=False)
        new._set_component("shape", tuple(new_shape), copy=False)

        return new

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return (
            f"<CF {self.__class__.__name__}{self.shape}: "
            f"{self}{self.original_shape}>"
        )

    @property
    def __in_memory__(self):
        """True if the array data is in memory.

        .. versionadded:: (cfdm) 1.11.2.0

        :Returns:

            `False`

        """
        return False

    def _get_array(self, index=None):
        """Returns a subspace of the dataset variable.

        The subspace is defined by the `index` attributes, and is
        applied with `cfdm.netcdf_indexer`.

        .. versionadded:: (cfdm) 1.11.2.0

        .. seealso:: `__array__`, `index`

        :Parameters:

            {{index: `tuple` or `None`, optional}}

        :Returns:

            `numpy.ndarray`
                The subspace.

        """
        return NotImplementedError(
            f"Must implement {self.__class__.__name__}._get_array"
        )

    @property
    def array(self):
        """Return an independent numpy array containing the data.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `numpy.ndarray`
                An independent numpy array of the data.

        **Examples**

        >>> n = numpy.asanyarray(a)
        >>> isinstance(n, numpy.ndarray)
        True

        """
        return self.__array__()

    def index(self, conform=True):
        """The index to be applied when converting to a `numpy` array.

        The `shape` is defined by the `index` applied to the
        `original_shape`.

        .. versionadded:: (cfdm) 1.11.2.0

        .. seealso:: `shape`, `original_shape`, `reference_shape`

        :Parameters:

            conform: `bool`, optional
                If True, the default, then

                * Convert a decreasing size 1 slice to an increasing
                  one.

                * Convert, where possible, a sequence of integers to a
                  slice.

                These transformations are to allow subspacing on data
                objects that have restricted indexing functionality,
                such as `h5py.Variable` objects.

                If False then these transformations are not done.

        :Returns:

            `tuple`

        **Examples**

        >>> x.shape
        (12, 145, 192)
        >>> x.index()
        (slice(None), slice(None), slice(None))
        >>> x = x[8:7:-1, 10:19:3, [15, 1,  4, 12]]
        >>> x = x[[0], [True, False, True], ::-2]
        >>> x.shape
        (1, 2, 2)
        >>> x.index()
        (slice(8, 9, None), slice(10, 17, 6), slice(12, -1, -11))
        >>> x.index(conform=False)
        (array([8]), array([10, 16]), array([12,  1]))

        """
        ind = self._custom.get("index")
        if ind is None:
            # No indices have been applied yet, so define indices that
            # are equivalent to Ellipsis, and set the original shape.
            shape = self.shape
            ind = tuple([slice(0, n, 1) for n in shape])
            self._custom["index"] = ind
            self._custom["original_shape"] = shape
            self._custom["reference_shape"] = shape
            return ind

        if not conform:
            return ind

        # Still here? Then conform the indices by:
        #
        # 1) Converting decreasing size 1 slices to increasing
        #    ones. This helps when the parent class can't cope with
        #    decreasing slices.
        #
        # 2) Converting, where possible, sequences of integers to
        #    slices. This helps when the parent class can't cope with
        #    indices that are sequences of integers.
        ind = list(ind)
        for n, (i, size) in enumerate(zip(ind[:], self.original_shape)):
            if isinstance(i, slice):
                if size == 1:
                    start, _, step = i.indices(size)
                    if step and step < 0:
                        # Decreasing slices are not universally
                        # accepted (e.g. `h5py` doesn't like them),
                        # but we can convert them to increasing ones.
                        ind[n] = slice(start, start + 1)
            elif np.iterable(i):
                i = normalize_index((i,), (size,))[0]
                if i.size == 1:
                    # Convert a sequence of one integer into a slice
                    start = i.item()
                    ind[n] = slice(start, start + 1)
                else:
                    # Convert a sequence of two or more evenly spaced
                    # integers into a slice.
                    step = np.unique(np.diff(i))
                    if step.size == 1:
                        start, stop = i[[0, -1]]
                        if stop >= start:
                            stop += 1
                        elif stop:
                            stop = -1
                        else:
                            stop = None

                        ind[n] = slice(start, stop, step.item())

        return tuple(ind)

    @property
    def reference_shape(self):
        """The shape of the data in the file with added dimensions.

        This is the same as `original_shape`, but with added size 1
        dimensions if `index` has new dimensions added with index
        values of `numpy.newaxis`.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `index`, `shape`, `original_shape`

        """
        out = self._custom.get("reference_shape")
        if out is None:
            # No subspace has been defined yet
            out = self.original_shape
            self._custom["reference_shape"] = out

        return out

    @property
    def original_shape(self):
        """The shape of the data in the file.

        The `shape` is defined by the result of subspacing the data in
        its original shape with the indices given by `index`.

        .. versionadded:: (cfdm) 1.11.2.0

        .. seealso:: `index`, `shape`, `reference_shape`

        """
        out = self._custom.get("original_shape")
        if out is None:
            # No subspace has been defined yet
            out = self.shape
            self._custom["original_shape"] = out

        return out

    def is_subspace(self):
        """True if the index represents a subspace of the data.

        The presence of `numpy.newaxis` (i.e. added size 1 dimensions)
        in `index` will not, on their own, cause `is_subspace` to
        return `False`

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `index`, `shape`

        """
        newaxis = np.newaxis
        index = [ind for ind in self.index() if ind is not newaxis]
        return index != [slice(0, n, 1) for n in self.original_shape]



import datetime
import faulthandler
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm


class RaggedIndexedContiguousArrayTest(unittest.TestCase):
    """Unit test for the RaggedIndexedContiguousArray class."""

    r = cfdm.RaggedIndexedContiguousArray(
        compressed_array=cfdm.Data(
            [
                280.0,
                281.0,
                279.0,
                278.0,
                279.5,
                281.0,
                282.0,
                278.0,
                279.0,
                277.5,
            ]
        ),
        shape=(2, 2, 3),
        index_variable=cfdm.Index(data=[0, 0, 0, 0, 1, 1, 1, 1]),
        count_variable=cfdm.Count(data=[1, 3, 2, 2]),
    )

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or calls (those
        # without a 'verbose' option to do the same) e.g. to debug them, wrap
        # them (for methods, start-to-end internally) as follows:
        # cfdm.log_level('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_RaggedIndexedContiguousArray_to_memory(self):
        """Test `RaggedIndexedContiguousArray.to_memory`"""
        self.assertIsInstance(
            self.r.to_memory(), cfdm.RaggedIndexedContiguousArray
        )

    def test_RaggedIndexedContiguousArray_get_count(self):
        """Test `RaggedIndexedContiguousArray.get_count`"""
        r = self.r.copy()
        self.assertIsInstance(r.get_count(), cfdm.Count)
        r._del_component("count_variable")
        self.assertIsNone(r.get_count(None))

    def test_RaggedIndexedContiguousArray_get_index(self):
        """Test `RaggedIndexedContiguousArray.get_index`"""
        r = self.r.copy()
        self.assertIsInstance(r.get_index(), cfdm.Index)
        r._del_component("index_variable")
        self.assertIsNone(r.get_index(None))


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



import os
import re

from setuptools import find_packages, setup


def _read(fname):
    """Returns content of a file."""
    fpath = os.path.dirname(__file__)
    fpath = os.path.join(fpath, fname)
    with open(fpath, "r") as file_:
        return file_.read()


def _get_version():
    """Returns library version by inspecting core/__init__.py file."""
    return re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        _read("cfdm/core/__init__.py"),
        re.MULTILINE,
    ).group(1)


version = _get_version()
packages = ["cfdm"]

long_description = """The **cfdm** Python package is a complete reference implementation
of the `CF data model
<https://www.geosci-model-dev.net/10/4619/2017>`_ for CF-1.11, that
identifies the fundamental elements of the `CF conventions
<http://cfconventions.org/>`_ and shows how they relate to each other,
independently of the `netCDF
<https://www.unidata.ucar.edu/software/netcdf/>`_ encoding.

The central element defined by the CF data model is the **field
construct**, which corresponds to a CF-netCDF data variable with all
of its metadata.

A simple example of reading a field construct from a file and
inspecting it:

    >>> import cfdm
    >>> f = cfdm.read('file.nc')
    >>> f
    [<Field: air_temperature(time(12), latitude(64), longitude(128)) K>]
    >>> print(f[0])
    Field: air_temperature (ncvar%tas)
    ----------------------------------
    Data            : air_temperature(time(12), latitude(64), longitude(128)) K
    Cell methods    : time(12): mean (interval: 1.0 month)
    Dimension coords: time(12) = [0450-11-16 00:00:00, ..., 0451-10-16 12:00:00] noleap
                    : latitude(64) = [-87.8638, ..., 87.8638] degrees_north
                    : longitude(128) = [0.0, ..., 357.1875] degrees_east
                    : height(1) = [2.0] m

The **cfdm** package can

* read field and domain constructs from netCDF and CDL datasets with a choice of netCDF backends,
* be fully flexible with respect to HDF5 chunking,
* create new field and domain constructs in memory,
* write and append field and domain constructs to netCDF datasets on disk,
* read, write, and manipulate UGRID mesh topologies,
* read, write, and create coordinates defined by geometry cells,
* read and write netCDF4 string data-type variables,
* read, write, and create netCDF and CDL datasets containing hierarchical groups,
* inspect field and domain constructs,
* test whether two constructs are the same,
* modify field and domain construct metadata and data,
* create subspaces of field and domain constructs,
* incorporate, and create, metadata stored in external files, and
* read, write, and create data that have been compressed by convention
  (i.e. ragged or gathered arrays, or coordinate arrays compressed
  by subsampling), whilst presenting a view of the data in its
  uncompressed form.

Documentation
=============

https://ncas-cms.github.io/cfdm

Dask
====

From version 1.11.2.0 the `cfdm` package uses `Dask
<https://docs.dask.org>`_ for all of its data manipulations.

Tutorial
========

https://ncas-cms.github.io/cfdm/tutorial

Installation
============

https://ncas-cms.github.io/cfdm/installation

Command line utility
====================

During installation the `cfdump` command line tool is also installed,
which generates text descriptions of the field constructs contained
in a netCDF dataset.

Source code
===========

This project is hosted in a `GitHub repository
<https://github.com/NCAS-CMS/cfdm>`_ where you can access the most
up-to-date source."""

# Get dependencies
requirements = open("requirements.txt", "r")
install_requires = requirements.read().splitlines()

tests_require = (
    [
        "pytest",
        "pycodestyle",
        "coverage",
        "scipy",
    ],
)
extras_require = {
    "documentation": [
        "sphinx==2.4.5",
        "sphinx-copybutton",
        "sphinx-toggleprompt",
        "sphinxcontrib-spelling",
    ],
    "pre-commit hooks": [
        "pre-commit",
        "black",
        "docformatter",
        "flake8",
        "pydocstyle",
    ],
}

setup(
    name="cfdm",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    version=version,
    description="A Python reference implementation of the CF data model",
    author="David Hassell, Sadie Bartholomew",
    author_email="david.hassell@ncas.ac.uk",
    maintainer="David Hassell, Sadie Bartholomew",
    url="https://ncas-cms.github.io/cfdm",
    download_url="https://pypi.org/project/cfdm/#files",
    platforms=["Linux", "MacOS", "Windows"],
    license="MIT",
    keywords=[
        "cf",
        "netcdf",
        "data",
        "science",
        "oceanography",
        "meteorology",
        "climate",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
    packages=find_packages(),
    scripts=["scripts/cfdump"],
    python_requires=">=3.9",
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require,
    include_package_data=True,
)



import numpy as np

from ...abstract import Array


class Subarray(Array):
    """Abstract base class for a subarray of compressed array.

    A subarray describes a unique part of the uncompressed array.

    .. versionadded:: (cfdm) 1.10.0.0

    """

    def __init__(
        self,
        data=None,
        indices=None,
        shape=None,
        compressed_dimensions=None,
        source=None,
        copy=True,
        context_manager=None,
    ):
        """**Initialisation**

        :Parameters:

            data: array_like
                The full compressed array spanning all subarrays, from
                which the elements for this subarray are defined by
                the *indices*.

            indices: `tuple`
                The indices of *data* that define this subarray.

            shape: `tuple` of `int`
                The shape of the uncompressed subarray.

            {{init compressed_dimensions: `dict`}}

                *Parameter example:*
                  ``{0: (0, 1)}``

                *Parameter example:*
                  ``{0: (0, 1, 2)}``

                *Parameter example:*
                  ``{2: (2, 3)}``

                *Parameter example:*
                  ``{1: (1,)}``

                *Parameter example:*
                  ``{0: (0,), 2: (2,)}``

            {{init source: optional}}

            {{init copy: `bool`, optional}}

            context_manager: function, optional
                A context manager that provides a runtime context for
                the conversion of *data* to a `numpy` array.

        """
        super().__init__(source=source, copy=copy)

        if source is not None:
            try:
                data = source._get_component("data", None)
            except AttributeError:
                data = None

            try:
                indices = source._get_component("indices", None)
            except AttributeError:
                indices = None

            try:
                shape = source._get_component("shape", None)
            except AttributeError:
                shape = None

            try:
                compressed_dimensions = source.compressed_dimensions()
            except AttributeError:
                compressed_dimensions = {}

        elif compressed_dimensions is None:
            compressed_dimensions = {}
        else:
            compressed_dimensions = compressed_dimensions.copy()

        if data is not None:
            self._set_component("data", data, copy=copy)

        if indices is not None:
            self._set_component("indices", indices, copy=False)

        if shape is not None:
            self._set_component("shape", shape, copy=False)

        self._set_component(
            "compressed_dimensions", compressed_dimensions, copy=False
        )

        self._set_component("context_manager", context_manager, copy=False)

    def __getitem__(self, indices):
        """Return a subspace of the uncompressed subarray.

        x.__getitem__(indices) <==> x[indices]

        Returns a subspace of the uncompressed subarray as an
        independent numpy array.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.__getitem__"
        )  # pragma: no cover

    def _asanyarray(self, data, indices=None, check_mask=True):
        """Convert data to a `numpy` array.

        The conversion takes place with a runtime context, if one has
        been provided.

        By default, the returned array will only a masked array if the
        data contains missing values.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            data: array_like
                The data to be converted.

            check_mask: `bool`, optional
                Check for masked elements, and if there are none
                convert the output to a non-masked `numpy` array.

        :Returns:

            `numpy.ndarray`
                The converted data.

        """
        context_manager = self._get_component("context_manager")
        if context_manager:
            # Convert the data to a numpy array within the given
            # runtime context
            with context_manager():
                if indices is not None:
                    data = data[indices]

                data = np.asanyarray(data)
        else:
            if indices is not None:
                data = data[indices]

            data = np.asanyarray(data)

        if check_mask and np.ma.isMA(data) and not np.ma.is_masked(data):
            data = np.array(data)

        return data

    def _select_data(self, data=None, check_mask=True):
        """Select compressed elements that correspond to this subarray.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            data: array_like or `None`
                A full compressed array spanning all subarrays, from
                which the elements for this subarray will be
                returned. By default, or if `None` then the `data`
                array is used.

            check_mask: `bool`, optional
                Check for masked elements in the selected data, and if
                there are none convert the output to a non-masked
                `numpy` array.

        :Returns:

            `numpy.ndarray`
                Values of the compressed array that correspond to this
                subarray.

        """
        if data is None:
            data = self.data

        return self._asanyarray(
            data, indices=self.indices, check_mask=check_mask
        )

    @property
    def data(self):
        """The full compressed array spanning all subarrays.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        return self._get_component("data")

    @property
    def dtype(self):
        """The data-type of the uncompressed data.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.dtype"
        )  # pragma: no cover

    @property
    def indices(self):
        """The indices of the data that define this subarray.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        return self._get_component("indices")

    @property
    def shape(self):
        """The shape of the uncompressed data.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        return self._get_component("shape")

    def compressed_dimensions(self):
        """Mapping of compressed to uncompressed dimensions.

        A dictionary key is a position of a dimension in the
        compressed data, with a value of the positions of the
        corresponding dimensions in the uncompressed data. Compressed
        array dimensions that are not compressed are omitted from the
        mapping.

        .. versionadded:: (cfdm) 1.10.0.0

        :Returns:

            `dict`
                The mapping of dimensions of the compressed array to
                their corresponding dimensions in the uncompressed
                array. Compressed array dimensions that are not
                compressed are omitted from the mapping.

        """
        c = self._get_component("compressed_dimensions")
        if not c:
            raise ValueError("No compressed dimensions have been defined")

        return c.copy()

    def get_filename(self, normalise=True, default=AttributeError()):
        """Return the name of the file containing the data.

        .. versionadded:: (cfdm) 1.10.0.2

        :Parameters:

            {{normalise: `bool`, optional}}

                .. versionadded:: (cfdm) 1.12.0.0

            default: optional
                Return the value of the *default* parameter if there
                is no file name.

                {{default Exception}}

                .. versionadded:: (cfdm) 1.12.0.0

        :Returns:

            `str`
                The file name.

        """
        filenames = self.data.get_filenames(normalise=normalise)
        if len(filenames) != 1:
            if default is None:
                return

            return self._default(
                default, f"{self.__class__.__name__} has no unique file name"
            )

        return filenames[0]



from copy import deepcopy
from os import sep
from os.path import join
from urllib.parse import urlparse

from s3fs import S3FileSystem
from uritools import isuri, urisplit

from ...functions import abspath, dirname
from . import Array


class FileArray(Array):
    """Abstract base class for an array in a file.

    .. versionadded:: (cfdm) 1.12.0.0

    """

    def __init__(
        self,
        filename=None,
        address=None,
        dtype=None,
        shape=None,
        mask=True,
        unpack=True,
        attributes=None,
        storage_options=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            filename: (sequence of `str`), optional
                The locations of datasets containing the array.

            address: (sequence of `str`), optional
                How to find the array in the datasets.

            dtype: `numpy.dtype`, optional
                The data type of the array. May be `None` if is not
                known. This may differ from the data type of the
                array in the datasets.

            shape: `tuple`, optional
                The shape of the dataset array.

            {{init mask: `bool`, optional}}

            {{init unpack: `bool`, optional}}

            {{init attributes: `dict` or `None`, optional}}

                If *attributes* is `None`, the default, then the
                attributes will be set during the first `__getitem__`
                call.

            {{init storage_options: `dict` or `None`, optional}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(source=source, copy=copy)

        if source is not None:
            try:
                shape = source._get_component("shape", None)
            except AttributeError:
                shape = None

            try:
                filename = source._get_component("filename", None)
            except AttributeError:
                filename = None

            try:
                address = source._get_component("address", None)
            except AttributeError:
                address = None

            try:
                dtype = source._get_component("dtype", None)
            except AttributeError:
                dtype = None

            try:
                mask = source._get_component("mask", True)
            except AttributeError:
                mask = True

            try:
                unpack = source._get_component("unpack", True)
            except AttributeError:
                unpack = True

            try:
                attributes = source._get_component("attributes", None)
            except AttributeError:
                attributes = None

            try:
                storage_options = source._get_component(
                    "storage_options", None
                )
            except AttributeError:
                storage_options = None

        if shape is not None:
            self._set_component("shape", shape, copy=False)

        if filename is not None:
            self._set_component("filename", filename, copy=False)

        if address is not None:
            self._set_component("address", address, copy=False)

        self._set_component("dtype", dtype, copy=False)
        self._set_component("mask", bool(mask), copy=False)
        self._set_component("unpack", bool(unpack), copy=False)

        if storage_options is not None:
            self._set_component("storage_options", storage_options, copy=copy)

        if attributes is not None:
            self._set_component("attributes", attributes, copy=copy)

        # By default, close the netCDF file after data array access
        self._set_component("close", True, copy=False)

    def __getitem__(self, indices):
        """Return a subspace of the array.

        x.__getitem__(indices) <==> x[indices]

        Returns a subspace of the array as an independent numpy array.

        """
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.__getitem__"
        )  # pragma: no cover

    def __repr__(self):
        """Called by the `repr` built-in function.

        x.__repr__() <==> repr(x)

        """
        return f"<CF {self.__class__.__name__}{self.shape}: {self}>"

    def __str__(self):
        """Called by the `str` built-in function.

        x.__str__() <==> str(x)

        """
        return f"{self.get_filename()}, {self.get_address()}"

    def __dask_tokenize__(self):
        """Return a value fully representative of the object.

        .. versionadded:: (cfdm) 1.12.0.0

        """
        return (
            self.__class__,
            self.shape,
            self.get_filename(normalise=True, default=None),
            self.get_address(),
            self.get_mask(),
            self.get_unpack(),
            self.get_attributes(copy=False),
            self.get_storage_options(),
        )

    def _get_array(self, index=None):
        """Returns a subspace of the dataset variable.

        The subspace is defined by the `index` attributes, and is
        applied with `cfdm.netcdf_indexer`.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `__array__`, `index`

        :Parameters:

            {{index: `tuple` or `None`, optional}}

        :Returns:

            `numpy.ndarray`
                The subspace.

        """
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}._get_array"
        )  # pragma: no cover

    @property
    def array(self):
        """Return an independent numpy array containing the data.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `numpy.ndarray`
                An independent numpy array of the data.

        **Examples**

        >>> n = {{package}}.{{class}}.array(a)
        >>> isinstance(n, numpy.ndarray)
        True

        """
        return self[...]

    @property
    def dtype(self):
        """Data-type of the array."""
        return self._get_component("dtype")

    @property
    def shape(self):
        """Shape of the array."""
        return self._get_component("shape")

    def close(self, dataset):
        """Close the dataset containing the data."""
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.close"
        )  # pragma: no cover

    def get_address(self, default=AttributeError()):
        """The name of the file containing the array.

        If there are multiple files then an `AttributeError` is
        raised by default.

        .. versionadded:: (cfdm) 1.10.1.0

        :Parameters:

            default: optional
                Return the value of the *default* parameter if there
                is no file.

                {{default Exception}}

        :Returns:

            `str`
                The file name.

        """
        address = self._get_component("address", None)
        if address is None:
            if default is None:
                return

            return self._default(
                default, f"{self.__class__.__name__} has no address"
            )

        return address

    def file_directory(self, normalise=False, default=AttributeError()):
        """The file directory.

        .. versionadded:: (cfdm) 1.12.0.0

        :Parameters:

            {{normalise: `bool`, optional}}

        :Returns:

            `str`
                The file directory name.

        **Examples**

        >>> a.get_filename()
        '/data1/file1'

        """
        filename = self.get_filename(normalise=normalise, default=None)
        if filename is None:
            if default is None:
                return

            return self._default(
                default, f"{self.__class__.__name__} has no file name"
            )

        return dirname(filename)

    def get_filename(self, normalise=False, default=AttributeError()):
        """The name of the file containing the array.

        .. versionadded:: (cfdm) 1.10.0.2

        :Parameters:

            {{normalise: `bool`, optional}}

                .. versionadded:: (cfdm) 1.12.0.0

            default: optional
                Return the value of the *default* parameter if there
                is no file name.

                {{default Exception}}

        :Returns:

            `str`
                The file name.

        """
        filename = self._get_component("filename", None)
        if filename is None:
            if default is None:
                return

            return self._default(
                default, f"{self.__class__.__name__} has no file name"
            )

        if normalise:
            filename = abspath(filename)

        return filename

    def get_mask(self):
        """Whether or not to automatically mask the data.

        .. versionadded:: (cfdm) 1.8.2

        **Examples**

        >>> b = a.get_mask()

        """
        return self._get_component("mask")

    def get_storage_options(
        self, create_endpoint_url=True, filename=None, parsed_filename=None
    ):
        """Return `s3fs.S3FileSystem` options for accessing S3 files.

        .. versionadded:: (cfdm) 1.12.0.0

        :Parameters:

            create_endpoint_url: `bool`, optional
                If True, the default, then create an
                ``'endpoint_url'`` option if and only if one was not
                set during object initialisation. In this case the
                ``'endpoint_url'`` will be set from the file name
                returned by `get_filename`, unless either of the
                *filename* or *parsed_filename* parameters is also
                set.

            filename: `str`, optional
                Used to set the ``'endpoint_url'`` if it was not
                set during object initialisation and
                *create_endpoint_url* is True. Ignored if the
                *parsed_filename* parameter has been set.

            parsed_filename: `urllib.parse.ParseResult`, optional
                Used to set the ``'endpoint_url'`` if it was not
                set during object initialisation and
                *create_endpoint_url* is True. Ignored if the
                *filename* parameter has been set.

        :Returns:

            `dict` or `None`
                The `s3fs.S3FileSystem` options.

        **Examples**

        >>> f.get_filename()
        's3://store/data/file.nc'
        >>> f.get_storage_options(create_endpoint_url=False)
        {}
        >>> f.get_storage_options()
        {'endpoint_url': 'https://store'}
        >>> f.get_storage_options(filename='s3://other-store/data/file.nc')
        {'endpoint_url': 'https://other-store'}
        >>> f.get_storage_options(create_endpoint_url=False,
        ...                       filename='s3://other-store/data/file.nc')
        {}

        >>> f.get_storage_options()
        {'key': 'scaleway-api-key...',
         'secret': 'scaleway-secretkey...',
         'endpoint_url': 'https://s3.fr-par.scw.cloud',
         'client_kwargs': {'region_name': 'fr-par'}}

        """
        storage_options = self._get_component("storage_options", None)
        if not storage_options:
            storage_options = {}
        else:
            storage_options = deepcopy(storage_options)

        client_kwargs = storage_options.get("client_kwargs", {})
        if (
            create_endpoint_url
            and "endpoint_url" not in storage_options
            and "endpoint_url" not in client_kwargs
        ):
            if parsed_filename is None:
                if filename is None:
                    try:
                        filename = self.get_filename(normalise=False)
                    except AttributeError:
                        pass
                    else:
                        parsed_filename = urlparse(filename)
                else:
                    parsed_filename = urlparse(filename)

            if parsed_filename is not None and parsed_filename.scheme == "s3":
                # Derive endpoint_url from filename
                storage_options["endpoint_url"] = (
                    f"https://{parsed_filename.netloc}"
                )

        return storage_options

    def open(self, func, *args, **kwargs):
        """Return a dataset file object and address.

        When multiple files have been provided an attempt is made to
        open each one, in the order stored, and a file object is
        returned from the first file that exists.

        .. versionadded:: (cfdm) 1.10.1.0

        :Parameters:

            func: callable
                Function that opens a file.

            args, kwargs: optional
                Optional arguments to *func*.

        :Returns:

            2-`tuple`
                The file object for the dataset, and the address of
                the data within the file.

        """
        filename = self.get_filename(normalise=True)
        url = urlparse(filename)
        if url.scheme == "file":
            # Convert a file URI into an absolute path
            filename = url.path
        elif url.scheme == "s3":
            # Create an openable S3 file object
            storage_options = self.get_storage_options(
                create_endpoint_url=True, parsed_filename=url
            )
            fs = S3FileSystem(**storage_options)
            filename = fs.open(url.path[1:], "rb")

        try:
            dataset = func(filename, *args, **kwargs)
        except FileNotFoundError:
            raise FileNotFoundError(f"No such file: {filename}")
        except RuntimeError as error:
            raise RuntimeError(f"{error}: {filename}")

        # Successfully opened a dataset, so return.
        return dataset, self.get_address()

    def replace_directory(self, old=None, new=None, normalise=False):
        """Replace the file directory.

        Modifies the name of the file.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `file_directory`, `get_filename`

        :Parameters:

            {{replace old: `str` or `None`, optional}}

            {{replace new: `str` or `None`, optional}}

            {{replace normalise: `bool`, optional}}

        :Returns:

            `{{class}}`
                A new `{{class}}` with modified file locations.

        **Examples**

        >>> a.get_filename()
        '/data/file1.nc'
        >>> b = a.replace_directory('/data', '/new/data/path/')
        >>> b.get_filename()
        '/new/data/path/file1.nc'
        >>> c = b.replace_directory('/new/data', None)
        >>> c.get_filename()
        'path/file1.nc'
        >>> c = b.replace_directory('path', '../new_path', normalise=False)
        >>> c.get_filename()
        '../new_path/file1.nc'
        >>> c = b.replace_directory(None, '/data')
        >>> c.get_filename()
        '/data/../new_path/file1.nc'
        >>> c = b.replace_directory('/new_path/', None, normalise=True)
        >>> c.get_filename()
        'file1.nc'

        """
        a = self.copy()

        filename = a.get_filename(normalise=normalise)
        if old or new:
            if normalise:
                if not old:
                    raise ValueError(
                        "When 'normalise' is True and 'new' is a non-empty "
                        "string, 'old' must also be a non-empty string."
                    )

                uri = isuri(filename)
                try:
                    old = dirname(old, normalise=True, uri=uri, isdir=True)
                except ValueError:
                    old = dirname(old, normalise=True, isdir=True)

                u = urisplit(old)
                if not uri and u.scheme == "file":
                    old = u.getpath()

                if new:
                    try:
                        new = dirname(new, normalise=True, uri=uri, isdir=True)
                    except ValueError:
                        new = dirname(new, normalise=True, isdir=True)

            if old:
                if filename.startswith(old):
                    if not new:
                        new = ""
                        if old and not old.endswith(sep):
                            old += sep

                    filename = filename.replace(old, new)
            elif new:
                if filename.startswith(sep):
                    filename = filename[1:]

                filename = join(new, filename)

        a._set_component("filename", filename, copy=False)
        return a

    def get_missing_values(self):
        """The missing values of the data.

        Deprecated at version 1.12.0.0. Use `get_attributes` instead.

        """

        class DeprecationError(Exception):
            """Deprecation error."""

            pass

        raise DeprecationError(
            f"{self.__class__.__name__}.get_missing_values was deprecated "
            "at version 1.12.0.0 and is no longer available. "
            f"Use {self.__class__.__name__}.get_attributes instead."
        )  # pragma: no cover

    def to_memory(self):
        """Bring data on disk into memory.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `numpy.ndarray`
                The new array.

        """
        return self.array

    def _set_attributes(self, var):
        """Set the netCDF variable attributes.

        These are set from the netCDF variable attributes, but only if
        they have not already been defined, either during {{class}}
        instantiation or by a previous call to `_set_attributes`.

        .. versionadded:: (cfdm) 1.12.0.0

        :Parameters:

            var: `netCDF4.Variable` or `h5netcdf.Variable`
                The netCDF variable.

        :Returns:

            `dict`
                The attributes.

        """
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}._set_attributes"
        )  # pragma: no cover

    def get_unpack(self):
        """Whether or not to automatically unpack the data.

        .. versionadded:: (cfdm) 1.12.0.0

        **Examples**

        >>> a.get_unpack()
        True

        """
        return self._get_component("unpack")

    def replace_filename(self, filename):
        """Replace the file location.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `file_directory`, `get_filename`,
                     `replace_directory`

        :Parameters:

            filename: `str`
                The new file location.

        :Returns:

            `{{class}}`
                A new `{{class}}` with modified file name.

        """
        a = self.copy()
        a._set_component("filename", filename, copy=False)
        return a



import logging

from . import core, mixin
from .decorators import _manage_log_level_via_verbosity

logger = logging.getLogger(__name__)


class CellMeasure(
    mixin.NetCDFVariable,
    mixin.NetCDFExternal,
    mixin.PropertiesData,
    mixin.Files,
    core.CellMeasure,
):
    """A cell measure construct of the CF data model.

    A cell measure construct provides information that is needed about
    the size or shape of the cells and that depends on a subset of the
    domain axis constructs. Cell measure constructs have to be used
    when the size or shape of the cells cannot be deduced from the
    dimension or auxiliary coordinate constructs without special
    knowledge that a generic application cannot be expected to have.

    The cell measure construct consists of a numeric array of the
    metric data which spans a subset of the domain axis constructs,
    and properties to describe the data. The cell measure construct
    specifies a "measure" to indicate which metric of the space it
    supplies, e.g. cell horizontal areas, and must have a units
    property consistent with the measure, e.g. square metres. It is
    assumed that the metric does not depend on axes of the domain
    which are not spanned by the array, along which the values are
    implicitly propagated. CF-netCDF cell measure variables correspond
    to cell measure constructs.

    **NetCDF interface**

    {{netCDF variable}}

    {{netCDF global attributes}}

    {{netCDF HDF5 chunks}}

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(
        self,
        measure=None,
        properties=None,
        data=None,
        source=None,
        copy=True,
        _use_data=True,
    ):
        """**Initialisation**

        :Parameters:

            measure: `str`, optional
                Set the measure that indicates the metric given by
                the data array.

                The measure may also be set after initialisation with
                the `set_measure` method.

                *Parameter example:*
                  ``measure='area'``

            {{init properties: `dict`, optional}}

               *Parameter example:*
                 ``properties={'standard_name': 'cell_area'}``

            {{init data: data_like, optional}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            measure=measure,
            properties=properties,
            data=data,
            source=source,
            copy=copy,
            _use_data=_use_data,
        )

        self._initialise_netcdf(source)
        self._initialise_original_filenames(source)

    def creation_commands(
        self,
        representative_data=False,
        namespace=None,
        indent=0,
        string=True,
        name="c",
        data_name="data",
        header=True,
    ):
        """Returns the commands to create the cell measure construct.

        .. versionadded:: (cfdm) 1.8.7.0

        .. seealso:: `{{package}}.Data.creation_commands`,
                     `{{package}}.Field.creation_commands`

        :Parameters:

            {{representative_data: `bool`, optional}}

            {{namespace: `str`, optional}}

            {{indent: `int`, optional}}

            {{string: `bool`, optional}}

            {{name: `str`, optional}}

            {{data_name: `str`, optional}}

            {{header: `bool`, optional}}

        :Returns:

            {{returns creation_commands}}

        **Examples**

        >>> x = {{package}}.CellMeasure(
        ...     measure='area',
        ...     properties={'units': 'm2'}
        ... )
        >>> x.set_data([100345.5, 123432.3, 101556.8])
        >>> print(x.creation_commands(header=False))
        c = {{package}}.CellMeasure()
        c.set_properties({'units': 'm2'})
        data = {{package}}.Data([100345.5, 123432.3, 101556.8], units='m2', dtype='f8')
        c.set_data(data)
        c.set_measure('area')

        """
        out = super().creation_commands(
            representative_data=representative_data,
            indent=0,
            namespace=namespace,
            string=False,
            name=name,
            data_name=data_name,
            header=header,
        )

        measure = self.get_measure(None)
        if measure is not None:
            out.append(f"{name}.set_measure({measure!r})")

        if string:
            indent = " " * indent
            out[0] = indent + out[0]
            out = ("\n" + indent).join(out)

        return out

    def dump(
        self,
        display=True,
        _omit_properties=None,
        _key=None,
        _level=0,
        _title=None,
        _axes=None,
        _axis_names=None,
    ):
        """A full description of the cell measure construct.

        Returns a description of all properties, including those of
        components, and provides selected values of all data arrays.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default the description is printed.

        :Returns:

            {{returns dump}}

        """
        if _title is None:
            name = self.identity(default=self.get_property("units", ""))
            _title = f"Cell Measure: {name}"

        if self.nc_get_external():
            if not (self.has_data() or self.properties()):
                ncvar = self.nc_get_variable(None)
                if ncvar is not None:
                    ncvar = f"ncvar%{ncvar}"
                else:
                    ncvar = ""

                _title += f" (external variable: {ncvar})"

        return super().dump(
            display=display,
            _key=_key,
            _omit_properties=_omit_properties,
            _level=_level,
            _title=_title,
            _axes=_axes,
            _axis_names=_axis_names,
        )

    @_manage_log_level_via_verbosity
    def equals(
        self,
        other,
        rtol=None,
        atol=None,
        verbose=None,
        ignore_data_type=False,
        ignore_fill_value=False,
        ignore_properties=None,
        ignore_compression=True,
        ignore_type=False,
    ):
        """Whether two cell measure constructs are the same.

        Equality is strict by default. This means that:

        * the same descriptive properties must be present, with the
          same values and data types, and vector-valued properties
          must also have same the size and be element-wise equal (see
          the *ignore_properties* and *ignore_data_type* parameters),
          and

        ..

        * if there are data arrays then they must have same shape and
          data type, the same missing data mask, and be element-wise
          equal (see the *ignore_data_type* parameter).

        {{equals tolerance}}

        {{equals compression}}

        Any type of object may be tested but, in general, equality is
        only possible with another cell measure construct, or a
        subclass of one. See the *ignore_type* parameter.

        {{equals netCDF}}

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            other:
                The object to compare for equality.

            {{atol: number, optional}}

            {{rtol: number, optional}}

            {{ignore_fill_value: `bool`, optional}}

            {{verbose: `int` or `str` or `None`, optional}}

            {{ignore_properties: (sequence of) `str`, optional}}

            {{ignore_data_type: `bool`, optional}}

            {{ignore_compression: `bool`, optional}}

            {{ignore_type: `bool`, optional}}

        :Returns:

            `bool`
                Whether the two cell measure constructs are equal.

        **Examples**

        >>> c = {{package}}.CellMeasure()
        >>> c.set_properties({'units': 'm2'})
        >>> c.equals(c)
        True
        >>> c.equals(c.copy())
        True
        >>> c.equals('not a cell measure')
        False

        >>> m = c.copy()
        >>> m.set_property('units', 'km2')
        >>> c.equals(m)
        False
        >>> c.equals(m, verbose=3)
        False
        >>> # Logs: CellMeasure: Different 'units' property values: 'm2', 'km2'

        """
        if not super().equals(
            other,
            rtol=rtol,
            atol=atol,
            verbose=verbose,
            ignore_data_type=ignore_data_type,
            ignore_fill_value=ignore_fill_value,
            ignore_properties=ignore_properties,
            ignore_compression=ignore_compression,
            ignore_type=ignore_type,
        ):
            return False

        measure0 = self.get_measure(None)
        measure1 = other.get_measure(None)
        if measure0 != measure1:
            logger.info(
                f"{self.__class__.__name__}: Different measure "
                f"({measure0} != {measure1})"
            )
            return False

        return True

    def identity(self, default=""):
        """Return the canonical identity.

        By default the identity is the first found of the following:

        * The measure, preceded by ``'measure:'``.
        * The ``standard_name`` property.
        * The ``cf_role`` property, preceded by 'cf_role='.
        * The ``long_name`` property, preceded by 'long_name='.
        * The netCDF variable name, preceded by 'ncvar%'.
        * The value of the default parameter.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `identities`

        :Parameters:

            default: optional
                If no identity can be found then return the value of the
                default parameter.

        :Returns:

                The identity.

        **Examples**

        >>> f = {{package}}.example_field(1)
        >>> c = f.get_construct('cellmeasure0')
        >>> c.get_measure()
        'area'

        >>> c.properties()
        {'units': 'km2'}
        >>> c.nc_get_variable()
        'cell_measure'
        >>> c.identity(default='no identity')
        'measure:area'

        >>> c.del_measure()
        'area'
        >>> c.identity()
        'ncvar%cell_measure'
        >>> c.nc_del_variable()
        'cell_measure'
        >>> c.identity()
        ''
        >>> c.identity(default='no identity')
        'no identity'

        """
        n = self.get_measure(None)
        if n is not None:
            return f"measure:{n}"

        n = self.get_property("standard_name", None)
        if n is not None:
            return n

        for prop in ("cf_role", "long_name"):
            n = self.get_property(prop, None)
            if n is not None:
                return f"{prop}={n}"

        n = self.nc_get_variable(None)
        if n is not None:
            return f"ncvar%{n}"

        return default

    def identities(self, generator=False, **kwargs):
        """Return all possible identities.

        The identities comprise:

        * The measure property, preceded by ``'measure:'``.
        * The ``standard_name`` property.
        * All properties, preceded by the property name and a colon,
          e.g. ``'long_name:Air temperature'``.
        * The netCDF variable name, preceded by ``'ncvar%'``.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `identity`

        :Parameters:

            generator: `bool`, optional
                If True then return a generator for the identities,
                rather than a list.

                .. versionadded:: (cfdm) 1.8.9.0

            kwargs: optional
                Additional configuration parameters that may be used
                by subclasses.

                .. versionadded:: (cfdm) 1.8.9.0

        :Returns:

            `list` or generator
                The identities.

        **Examples**

        >>> f = {{package}}.example_field(1)
        >>> c = f.get_construct('cellmeasure0')
        >>> c.get_measure()
        'area'

        >>> c.properties()
        {'units': 'km2'}
        >>> c.nc_get_variable()
        'cell_measure'
        >>> c.identities()
        ['measure:area', 'units=km2', 'ncvar%cell_measure']
        >>> for i in c.identities(generator=True):
        ...     print(i)
        ...
        measure:area
        units=km2
        ncvar%cell_measure

        """
        measure = self.get_measure(None)
        if measure is not None:
            pre = ((f"measure:{measure}",),)
            pre0 = kwargs.pop("pre", None)
            if pre0:
                pre = tuple(pre0) + pre

            kwargs["pre"] = pre

        return super().identities(generator=generator, **kwargs)



from .bilinearsubarray import BiLinearSubarray
from .boundsfromnodessubarray import BoundsFromNodesSubarray
from .cellconnectivitysubarray import CellConnectivitySubarray
from .gatheredsubarray import GatheredSubarray
from .biquadraticlatitudelongitudesubarray import (
    BiQuadraticLatitudeLongitudeSubarray,
)
from .interpolationsubarray import InterpolationSubarray
from .linearsubarray import LinearSubarray
from .pointtopologyfromedgessubarray import PointTopologyFromEdgesSubarray
from .pointtopologyfromfacessubarray import PointTopologyFromFacesSubarray
from .quadraticlatitudelongitudesubarray import (
    QuadraticLatitudeLongitudeSubarray,
)
from .quadraticsubarray import QuadraticSubarray
from .raggedsubarray import RaggedSubarray



from ..docstring import _docstring_substitution_definitions
from ..functions import deepcopy
from ..meta import DocstringRewriteMeta


class Container(metaclass=DocstringRewriteMeta):
    """Abstract base class for storing components.

    .. warning:: The ``custom`` component dictionary is only shallow
                 copied when initialised from the *source* parameter,
                 regardless of the value of the *copy* parameter. This
                 is to avoid potentially expensive deep copies of the
                 dictionary values.

                 A subclass of `Container` that requires custom
                 dictionary values to be deep copied should ensure
                 that this occurs its `__init__` method.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(self, source=None, copy=True):
        """**Initialiation**

        :Parameters:

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        self._components = {}

        if source is not None:
            # WARNING: The 'custom' dictionary is only shallow copied
            #          from source
            try:
                custom = source._get_component("custom", {})
            except AttributeError:
                custom = {}
            else:
                custom = custom.copy()
        else:
            custom = {}

        self._set_component("custom", custom, copy=False)

    def __deepcopy__(self, memo):
        """Called by the `copy.deepcopy` function.

        x.__deepcopy__() <==> copy.deepcopy(x)

        .. versionadded:: (cfdm) 1.7.0

        **Examples**

        >>> import copy
        >>> f = {{package}}.{{class}}()
        >>> g = copy.deepcopy(f)

        """
        return self.copy()

    def __docstring_substitutions__(self):
        """Defines docstring substitutions for a class and subclasses.

        That is, defines docstring substitutions that apply to this
        class and all of its subclasses.

        These are in addtion to, and take precendence over, docstring
        substitutions defined by the base classes of this class.

        See `_docstring_substitutions` for details.

        .. versionaddedd:: (cfdm) 1.8.7.0

        :Returns:

            `dict`
                The docstring substitutions that have been applied.

        """
        return _docstring_substitution_definitions

    def __docstring_package_depth__(self):
        """Returns the package depth for {{package}} substitutions.

        See `_docstring_package_depth` for details.

        """
        return 1

    def _default(self, default, message=None):
        """Return a value or raise an Exception for a default case.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            default:
                The value to return, or to raise if set to an `Exception`
                instance.

            message: `str`, optional
                The error message to raise with *default* if it is an
                `Exception` instance.

        :Returns:

                The value of *default* if it is not an `Exception`
                instance.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f._default(AttributeError())  # Raises Exception
        Traceback (most recent call last):
            ...
        AttributeError
        >>> f._default(ValueError("Missing item"))  # Raises Exception
        Traceback (most recent call last):
            ...
        ValueError: Missing item
        >>> f._default(ValueError(), message="No component")  # Raises Exception
        Traceback (most recent call last):
            ...
        ValueError: No component
        >>> f._default(False)
        False
        >>> f._default('Not set')
        'Not set'
        >>> f._default(1)
        1

        """
        if isinstance(default, Exception):
            if message is not None and not default.args:
                default = type(default)(message)

            raise default

        return default

    def _del_component(self, component, default=ValueError()):
        """Remove a component.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `_get_component`, `_has_component`, `_set_component`

        :Parameters:

            component:
                The name of the component to be removed.

            default: optional
                Return *default* if the component has not been set.

                {{default Exception}}

        :Returns:

                The removed component. If unset then *default* is
                returned, if provided.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f._set_component('foo', 'bar')
        >>> f._has_component('foo')
        True
        >>> f._get_component('foo')
        'bar'
        >>> f._del_component('foo')
        'bar'
        >>> f._has_component('foo')
        False

        """
        try:
            return self._components.pop(component)
        except KeyError:
            if default is None:
                return

            return self._default(
                default,
                f"{self.__class__.__name__} has no {component!r} component",
            )

    @property
    def _custom(self):
        """Customisable storage for additional attributes.

        See https://ncas-cms.github.io/cfdm/extensions.html for more
        information.

        .. versionadded:: (cfdm) 1.7.4

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f._custom
        {}
        >>> f._custom['feature_1'] = 1
        >>> g = f.copy()
        >>> g._custom['feature_2'] = 2
        >>> f._custom
        {'feature_1': 1}
        >>> g._custom
        {'feature_1': 1, 'feature_2': 2}
        >>> del g._custom['feature_1']
        >>> g._custom
        {'feature_2': 2}
        >>> f._custom
        {'feature_1': 1}

        """
        return self._get_component("custom")

    def _get_component(self, component, default=ValueError()):
        """Return a component.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `_del_component`, `_has_component`, `_set_component`

        :Parameters:

            component:
                The name of the component to be returned.

            default: optional
                Return *default* if the component has not been set.

                {{default Exception}}

        :Returns:

                The component. If unset then *default* is returned, if
                provided.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f._set_component('foo', 'bar')
        >>> f._has_component('foo')
        True
        >>> f._get_component('foo')
        'bar'
        >>> f._del_component('foo')
        'bar'
        >>> f._has_component('foo')
        False

        """
        try:
            return self._components[component]
        except KeyError:
            if default is None:
                return

            return self._default(
                default,
                f"{self.__class__.__name__} has no {component!r} component",
            )

    def _has_component(self, component):
        """Whether a component has been set.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `_del_component`, `_get_component`, `_set_component`

        :Parameters:

            component:
                The name of the component.

        :Returns:

            `bool`
                True if the component has been set, otherwise False.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f._set_component('foo', 'bar')
        >>> f._has_component('foo')
        True
        >>> f._get_component('foo')
        'bar'
        >>> f._del_component('foo')
        'bar'
        >>> f._has_component('foo')
        False

        """
        return component in self._components

    def _set_component(self, component, value, copy=True):
        """Set a component.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `_del_component`, `_get_component`, `_has_component`

        :Parameters:

            component: `str`
                The name of the component.

            value:
                The value for the component.

        :Returns:

            `None`

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f._set_component('foo', 'bar')
        >>> f._has_component('foo')
        True
        >>> f._get_component('foo')
        'bar'
        >>> f._del_component('foo')
        'bar'
        >>> f._has_component('foo')
        False

        """
        if copy:
            value = deepcopy(value)

        self._components[component] = value

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def copy(self):
        """Return a deep copy.

        ``f.copy()`` is equivalent to ``copy.deepcopy(f)``.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `{{class}}`
                The deep copy.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> g = f.copy()

        """
        return type(self)(source=self, copy=True)
